try:
    import copy 
    import functools, time
    import numpy as np
    from scipy.spatial import Voronoi
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing copy: {str(e)}\n")
    del sys

try:
    from .BasePartition import BasePartition
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing BasePartition: {str(e)}\n")
    del sys

def timeit_decorator(func):
    """Decorator to measure the execution time of a function."""
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()  # Record the start time
        result = func(*args, **kwargs)  # Execute the function
        end_time = time.time()  # Record the end time
        execution_time = end_time - start_time  # Calculate the execution time
        print(f"Execution time of {func.__name__}: {execution_time} seconds")
        return result
    return wrapper

class CrystalDefect_builder(BasePartition):
    def __init__(self, *args, **kwargs):
        """
        Initializes the CrystalDefectBuilder object. If a Periodic_Object is provided, its attributes are copied.

        Parameters:
            file_location (str): Location of the file.
            name (str): Name of the object.
            Periodic_Object (object): An object containing periodic attributes.
            **kwargs: Additional keyword arguments.
        """
        super().__init__(*args, **kwargs)

    def generate_combinations(self, lists: list, counts: list):
        """
        Generates all possible combinations of elements from the given lists based on the specified counts.

        Parameters:
            lists (list of lists): List of lists containing elements.
            counts (list of int): List of counts specifying the number of elements to select from each list.

        Returns:
            list: A list of combinations.
        """
        if len(lists) != len(counts):
            raise ValueError("The number of lists must be equal to the number of counts")

        def combinations_helper(current_combination, remaining_lists, remaining_counts):
            if not remaining_lists:
                yield current_combination
                return

            next_list = remaining_lists[0]
            next_count = remaining_counts[0]

            for combination in combinations(next_list, next_count):
                yield from combinations_helper(current_combination + combination, remaining_lists[1:], remaining_counts[1:])

        def combinations(lst, count):
            if count == 0:
                yield []
                return
            if not lst:
                return

            for i in range(len(lst)):
                for tail in combinations(lst[i+1:], count-1):
                    yield [lst[i]] + tail

        result = list(combinations_helper([], lists, counts))
        return result
    
    @staticmethod
    def get_probability(combination: list = None, distribution: str = 'uniform', AtomPositionManager: object = None, distances_dict: dict = None, coefficient: float = 1.0) -> float:
        """
        Calculates the probability of a combination based on the specified distribution.

        Parameters:
            combination (list): The combination of atom IDs.
            distribution (str): The type of distribution ('uniform', 'distance', 'distance-1', 'random').
            AtomPositionManager (object): An object managing atom positions.
            distances_dict (dict): A dictionary containing precomputed distances.
            coefficient (float): A coefficient to modify the probability calculation.

        Returns:
            float: The calculated probability.
        """
        if distribution in ['random', 'uniform']:
            return 1

        if distribution == 'distance':
            if distances_dict:
                return np.sum([distances_dict[n2][n1] if n1 > n2 else distances_dict[n1][n2] for i, n1 in enumerate(combination) for j, n2 in enumerate(combination[i:])]) ** coefficient
            else:
                return np.sum([AtomPositionManager.distance(AtomPositionManager.atomPositions[n1], AtomPositionManager.atomPositions[n2]) for i, n1 in enumerate(combination) for j, n2 in enumerate(combination[i:])]) ** coefficient

        if distribution == 'distance-1':
            if distances_dict:
                return 1 / np.sum([distances_dict[n2][n1] if n1 > n2 else distances_dict[n1][n2] for i, n1 in enumerate(combination) for j, n2 in enumerate(combination[i:])]) ** coefficient
            else:
                return 1 / np.sum([AtomPositionManager.distance(AtomPositionManager.atomPositions[n1], AtomPositionManager.atomPositions[n2]) for i, n1 in enumerate(combination) for j, n2 in enumerate(combination[i:])]) ** coefficient

        return 0

    @staticmethod
    def get_atom_ids(AtomPositionManager, selection_list):
        """
        Retrieves atom IDs based on the selection list which can contain atom indices or labels.

        Parameters:
            AtomPositionManager (object): An object managing atom positions and labels.
            selection_list (list): A list of atom indices or labels.

        Returns:
            list: A list of unique atom IDs.
        """
        is_number = lambda s: s.replace('.', '', 1).isdigit() or s.replace('-', '', 1).replace('.', '', 1).isdigit()

        atom_ids = set()
        for item in selection_list:
            if isinstance(item, int) or is_number(item):
                atom_ids.add( int(item) )
            elif isinstance(item, str):

                if item.upper() in ['H2O', 'water']:
                    for i, label in enumerate(AtomPositionManager.atomLabelsList):
                        if label == 'O':
                            pos = AtomPositionManager.atomPositions[i]
                            closest_neighbors = AtomPositionManager.find_n_closest_neighbors(pos, 4)
                            if AtomPositionManager.atomLabelsList[closest_neighbors[1][1]] == AtomPositionManager.atomLabelsList[closest_neighbors[1][2]] == 'H' and closest_neighbors[0][2] < 1.2 and closest_neighbors[0][3] > 1.2: 
                                atom_ids.add( tuple(closest_neighbors[1][:3]) )
           
                if item.upper() in ['H-OH']:
                    for i, label in enumerate(AtomPositionManager.atomLabelsList):
                        if label == 'O':
                            pos = AtomPositionManager.atomPositions[i]
                            closest_neighbors = AtomPositionManager.find_n_closest_neighbors(pos, 4)
                            if AtomPositionManager.atomLabelsList[closest_neighbors[1][1]] == AtomPositionManager.atomLabelsList[closest_neighbors[1][2]] == 'H' and closest_neighbors[0][2] < 1.2 and closest_neighbors[0][3] > 1.2: 
                                atom_ids.update([closest_neighbors[1][1], closest_neighbors[1][2]])

                elif '-' in item:
                    a1, a2 = item.split('-')
                    for i, label in enumerate(AtomPositionManager.atomLabelsList):
                        if label == a1:
                            
                            pos = AtomPositionManager.atomPositions[i]
                            if AtomPositionManager.atomLabelsList[AtomPositionManager.find_n_closest_neighbors(pos, 2)[1][1]] == a2 or \
                                AtomPositionManager.atomLabelsList[AtomPositionManager.find_n_closest_neighbors(pos, 3)[1][2]] == a2 :
                                atom_ids.add( int(i) )

                else:
                    atom_ids.update([i for i, label in enumerate(AtomPositionManager.atomLabelsList) if label == item])
        
        return list(atom_ids) 

    @staticmethod
    def get_distances_dict(atom_groups_ID: np.array, AtomPositionManager: object) -> dict:
        """
        Creates a dictionary of distances between atom groups.

        Parameters:
            atom_groups_ID (np.array): An array of atom IDs.
            AtomPositionManager (object): An object managing atom positions.

        Returns:
            dict: A dictionary of distances between atom groups.
        """
        distances_dict = { n:{n:0} for n in atom_groups_ID }
        
        for n1 in atom_groups_ID:

            if isinstance(n1, (list, tuple, np.ndarray)):
                # If n1 is a list, tuple, or an array of indices, calculate the mean position.
                n1_positions = np.mean([AtomPositionManager.atomPositions[ap] for ap in n1], axis=0)
            elif isinstance(n1, (int, np.integer, np.int64)):
                # If n1 is a single integer value, use the position directly associated with that index.
                n1_positions = AtomPositionManager.atomPositions[n1]
            else:
                raise TypeError("n1 must be an integer, a list, a tuple, or a numpy array of integers")

            for n2 in atom_groups_ID:
                if n1 < n2:
                    if isinstance(n2, (list, tuple, np.ndarray)):
                        # If n2 is a list, tuple, or an array of indices, calculate the mean position.
                        n2_positions = np.mean([AtomPositionManager.atomPositions[ap] for ap in n2], axis=0)
                    elif isinstance(n2, (int, np.integer, np.int64)):
                        # If n2 is a single integer value, use the position directly associated with that index.
                        n2_positions = AtomPositionManager.atomPositions[n2]
                    else:
                        raise TypeError("n1 must be an integer, a list, a tuple, or a numpy array of integers")

                    distances_dict[n1][n2] = AtomPositionManager.distance( n1_positions, n2_positions )

        return distances_dict

    def generate_configurational_space(self, values, verbose:bool=False, containers:list=None):
        """
        Generates the configurational space by creating various combinations of defects in the crystal structure.

        Parameters:
            values (dict): A dictionary containing parameters for generating the configurational space.

        Returns:
            list: A list of containers with updated configurations.
        """
        iterations = values.get('iterations', 1)
        distribution = values.get('distribution', 'uniform')
        fill = values.get('fill', False)
        repetitions = values.get('repetitions', 1)
        atom_groups = values.get('atom_groups', None)
        group_numbers = values.get('group_numbers', None)
        coefficient = values.get('coefficient', 10)
        Nw = values.get('Nw', None)

        containers_new = [] # here we alocate new configurations 
        containers = containers if isinstance(containers, list) else self.containers

        if not group_numbers or (isinstance(group_numbers,list) and len(group_numbers) == 0):
            group_numbers = np.ones( len(atom_groups) )
        
        for container_index, container in enumerate(containers):
            if verbose: print(f'>> Analising container {container_index}')
            for rep in range(repetitions):
                if verbose: print(f' - Repetitions : {rep} ({100*float(rep)/repetitions}%)')    
                
                container_copy = self.copy_and_update_container(container, f'/generate_configurational_space/{container_index}_{rep}', '')
                
                for ite in range(iterations):
                    if verbose: print(f' - iteration : {ite} ({100*float(ite)/iterations}%)')    
                    
                    atom_groups_ID = [self.get_atom_ids(container_copy.AtomPositionManager, group) for group in atom_groups]
                    if verbose: print(f' - Processing : {len(atom_groups_ID)} groups : { [len(agi) for agi in atom_groups_ID] } groups len ')    

                    if 'distance' in distribution:
                        distances_dict = self.get_distances_dict(list(set(item for sublist in atom_groups_ID for item in sublist)), container_copy.AtomPositionManager) 
                        if verbose: print(f' - Distance dict calculated (ok!) ')    
                    else:
                        distances_dict = {}

                    combinations = self.generate_combinations(atom_groups_ID, group_numbers)
                    if verbose: print(f' - Founded {len(combinations)} ')    

                    probabilities = [self.get_probability(  combination=c, distribution=distribution, AtomPositionManager=container_copy.AtomPositionManager, 
                                                            distances_dict=distances_dict, coefficient=coefficient) for c in combinations]
                    if verbose: print(f' - Probabilities calculated')    

                    probabilities_norm = np.array(probabilities) / np.sum(probabilities)
                    if verbose: print(f' - Probabilities normalized')

                    if len(combinations) == 0:
                        continue
                    selected_index = np.random.choice(np.arange(len(combinations)), p=probabilities_norm)
                    selected_combination = np.array(combinations[selected_index]).flatten()

                    container_copy.AtomPositionManager.atomLabelsList = np.array(container_copy.AtomPositionManager.atomLabelsList)

                    if fill:
                        #translation = np.mean([container_copy.AtomPositionManager.atomPositions[n] for n in selected_combination], axis=0)
                        translation = np.sum([container_copy.AtomPositionManager.atomPositions[n]*container_copy.AtomPositionManager.covalent_radii[container_copy.AtomPositionManager.atomLabelsList[n]]**2 for n in selected_combination]) / np.sum([container_copy.AtomPositionManager.covalent_radii[container_copy.AtomPositionManager.atomLabelsList[n]]**2 for n in selected_combination])
                        r = 2 + np.max([container_copy.AtomPositionManager.distance(container_copy.AtomPositionManager.atomPositions[n1], container_copy.AtomPositionManager.atomPositions[n2]) for i, n1 in enumerate(combinations[selected_index]) for j, n2 in enumerate(combinations[selected_index][i:])]) 

                        # mean distance for N point in a r radius circle
                        if verbose: print(f' - Fill with water | sphere r={r} c={translation}')

                    container_copy.AtomPositionManager.remove_atom(selected_combination)
                    
                    if fill:
                        values = {
                            'density': 1,
                            'solvent': ['H2O'],
                            'molecules_number': {'H2O':Nw},
                            'max_iteration':100000,
                            'shape': 'sphere',
                            'size': [r],
                            'colition_tolerance': 1.7,
                            'translation': translation,
                            'distribution':'center',
                            'wrap': True,
                            'verbose':verbose
                        }

                        container_solvent = self.handleCLUSTER(values= {'ADD_SOLVENT':values}, containers=[container_copy])
                        
                        while container_solvent == False:
                            values['max_iteration'] += 5000
                            values['colition_tolerance'] -= .02
                            values['size'][0] -= .1

                            container_solvent = self.handleCLUSTER(values= {'ADD_SOLVENT':values}, containers=[container_copy])
                        
                        container_copy = container_solvent[0]

                containers_new.append(container_copy)
                if verbose: print(f' - Configuration generated (ok!)')
            

        self.containers = containers_new

        return containers_new

    def handleVacancy(self, container, values, container_index, file_location=None):
        sub_directories, containers = [], []

        container_copy = self.copy_and_update_container(container, '/Vacancy', file_location)
        container_copy.AtomPositionManager = CrystalDefectGenerator(Periodic_Object=container_copy.AtomPositionManager)
        all_vacancy_configs, all_vacancy_label = container_copy.AtomPositionManager.generate_all_vacancies()

        for cv_i, (vacancy_configs, vacancy_label) in enumerate(zip(all_vacancy_configs, all_vacancy_label)):
            container_copy2 = copy.deepcopy(container_copy)
            container_copy2.AtomPositionManager = vacancy_configs
            container_copy2.file_location = f'{container_copy.file_location}/{cv_i}_{vacancy_label}'
            sub_directories.append(f'{cv_i}_{vacancy_label}')
            containers.append(container_copy2)
        
        self.generate_execution_script_for_each_container(sub_directories, container.file_location + '/Vacancy')
        return containers

    def handleDefect(self, container, values, container_index, file_location):

        for v in values:
            if v['defect'].upper() == 'SUBSTITUTION':
                self.generate_all_substitutional_impurity( container=container, species=v['species'], new_species=v['new_species'], weights=v['weights'] )

    # ------------------------------------------------------------------------------------#
    def _generate_defect_configurations(self, defect_introducer, configurations:list=None):
        """
        General method to generate defect configurations.

        Parameters:
        - defect_introducer (function): Function that introduces the specific defect.
        - positions (list or None): Positions to introduce defects.
        - labels (list or None): Labels of atoms to introduce defects.

        Returns:
        - tuple: Two lists containing defect configurations and corresponding labels.
        """
        all_configs = []
        all_labels = []

        for config in configurations:
            temp_manager = copy.deepcopy(self)
            method = getattr(temp_manager, defect_introducer, None)

            if callable(method):
                method(**config)
            else:
                print(f"ERROR '{defect_introducer}' does not exist.")

            if not self._is_redundant(all_configs, temp_manager):
                all_configs.append(temp_manager)
                all_labels.append( '_'.join([str(c) for c in config.values()]) )

        return all_configs, all_labels


    # ------------------------------------------------------------------------------------ #
    def generate_all_vacancies(self, atomlabel=None):
        """
        Generate all possible vacancy configurations for the system.

        Parameters:
        - atomlabel (list or None): Specifies the type of atom for which vacancies should be generated.

        Returns:
        - tuple: Two lists containing vacancy configurations and corresponding labels.
        """
        # Parameters:
        atomlabel = list(atomlabel) if atomlabel is str else atomlabel

        # Determine the indices at which vacancies should be introduced
        if atomlabel:   indices = [i for i, label in enumerate(self.atomLabelsList) if label in atomlabel]
        else:           indices = np.array(range(self._atomCount), dtype=np.int64)

        configurations = [ {'atom_index':i} for i in indices  ]
        return self._generate_defect_configurations('introduce_vacancy', configurations)
    
    def generate_all_interstitial(self, atomlabel:list, new_atom_position:np.array=None):
        """

        """
        # Parameters: 
        new_atom_position = new_atom_position if new_atom_position is not None else self._find_volumes_center()
        new_atom_position = [nap for nap in new_atom_position if self.is_point_inside_unit_cell(nap) ] 

        # Determine the indices at which vacancies should be introduced
        if atomlabel:   indices = [i for i, label in enumerate(self.atomLabelsList) if label in atomlabel]
        else:           indices = np.array(range(self._atomCount), dtype=np.int64)

        configurations = [ {'new_atom_label':al, 'new_atom_position':nap } for al in atomlabel for nap in new_atom_position ]
        return self._generate_defect_configurations('introduce_interstitial', configurations)
    
    def generate_all_self_interstitial(self, atomlabel:list, new_atom_position:np.array=None):
        """

        """
        return self.generate_all_interstitial(atomlabel=self.uniqueAtomLabels, new_atom_position=None)
    
    def generate_one_substitutional_impurity(self, container, species:list, new_species:list):
        '''
        md = MBTR.MDTR(lattice_vectors=container.AtomPositionManager.latticeVectors, 
                    atomLabelsList=container.AtomPositionManager.atomLabelsList, 
                    atomPositions=container.AtomPositionManager.atomPositions, )

        rep, rep_div = md.get_mdtr()
        similarity_matrix = md.get_selfsimilarity_matrix( np.sum( np.abs( rep_div[0,:,:,:]), axis=2).T) 

        groups = md.find_related_atoms_groups(similarity_matrix, threshold=0.82)
        '''
        from ...descriptor import MBTR

        ans = []
        for i, ID in enumerate(container.AtomPositionManager.atomLabelsList):
            if ID in species:
                for ID2 in new_species:
                    container_copy = copy.deepcopy(container)
                    container_copy.AtomPositionManager.set_ID( atom_index=i, ID=ID2 )
                    ans.append(container_copy)

                    md = MBTR.MDTR(lattice_vectors=container_copy.AtomPositionManager.latticeVectors, 
                                atomLabelsList=container_copy.AtomPositionManager.atomLabelsList, 
                                atomPositions=container_copy.AtomPositionManager.atomPositions, )

                    rep, rep_div = md.get_mdtr()
                    #similarity_matrix = md.get_selfsimilarity_matrix( np.sum( np.abs( rep_div[0,:,:,:]), axis=2).T) 

                    #groups = md.find_related_atoms_groups(similarity_matrix, threshold=0.82)
                    ans1.append( np.abs(rep[0,:]) )
                    plt.plot( np.abs(rep[0,:]) )
                    
    def generate_all_substitutional_impurity(self, container, species:list, new_species:list, ):
        """

        """
        nes = self.generate_one_substitutional_impurity(container=container, species=species, new_species=new_species)

        '''
        md = MBTR.MDTR(lattice_vectors=self.containers[0].AtomPositionManager.latticeVectors, 
                    atomLabelsList=self.containers[0].AtomPositionManager.atomLabelsList, 
                    atomPositions=self.containers[0].AtomPositionManager.atomPositions, )

        rep, rep_div = md.get_mdtr()
        similarity_matrix = md.get_selfsimilarity_matrix( np.sum( np.abs( rep_div[0,:,:,:]), axis=2).T) 

        groups = md.find_related_atoms_groups(similarity_matrix, threshold=0.82)
        print(groups)
        print( self.containers[0].AtomPositionManager.atomPositions[4,:] )
        print( self.containers[0].AtomPositionManager.atomPositions[5,:] )
        print( self.containers[0].AtomPositionManager.atomPositions[10,:] )
        print(rep.shape , rep_div.shape) 
        print('dif', np.sum( (np.abs(rep_div[0,:,13,0].T) - np.abs(rep_div[0,:,17,0].T))**2 )**0.5 )
        print('dif', np.sum( (np.abs(rep_div[0,:,0,0].T) - np.abs(rep_div[0,:,1,0].T))**2 )**0.5 )
        print('dif', np.sum( (np.abs(rep_div[0,:,4,0].T) - np.abs(rep_div[0,:,5,0].T))**2 )**0.5 )
        print('dif', np.sum( (np.abs(rep_div[0,:,12,0].T) - np.abs(rep_div[0,:,13,0].T))**2 )**0.5 )
        mask = self.containers[0].AtomPositionManager.atomLabelsList == 'Ni'
                
        plt.plot( np.abs(rep_div[0,:,4,1].T) )
        plt.plot( np.abs(rep_div[0,:,5,1].T) )
        plt.plot( np.abs(rep_div[0,:,10,2].T) )
        #similarity_matrix[similarity_matrix>0.8] = 1
        #similarity_matrix[similarity_matrix<=0.8] = 0
        plt.matshow( similarity_matrix )

        plt.show()
        print(a.shape, b.shape)

        asdf
        '''
        '''
        # Parameters: 
        atomlabel = atomlabel if atomlabel is not None else self.uniqueAtomLabels
        new_atom_label = list(new_atom_label) if new_atom_label is list else new_atom_label
        # Determine the indices at which vacancies should be introduced
        if atomlabel:   indices = [i for i, label in enumerate(self.atomLabelsList) if label in atomlabel]
        else:           indices = np.array(range(self._atomCount), dtype=np.int64)
        print(indices)

        configurations = [ {'atom_index':i, 'new_atom_label':nal } for i in indices for nal in new_atom_label ]
        return self._generate_defect_configurations('introduce_substitutional_impurity', configurations)

        '''

    def _find_volumes_center(self, atomPositions:np.array=None):
        """
        Finds potential volumes for new atoms in a structure.

        Args:
            atom_coordinates (list of list of floats): List of existing atom coordinates.

        Returns:
            list of Voronoi region vertices: List of vertices of the Voronoi regions.
        """
        # Convert coordinates to a NumPy array.
        atomPositions = atomPositions if atomPositions is not None else self.atomPositions

        # Calculate the Voronoi decomposition.
        vor = Voronoi(atomPositions)

        return vor.vertices


'''
path = '/home/akaris/Documents/code/Physics/VASP/v6.1/files/dataset/CoFeNiOOH_jingzhu/surf_CoFe_2H_4OH/vacancy'
ap = CrystalDefectGenerator(file_location=path+'/POSCAR')
ap.readPOSCAR()
all_configs, all_labels = ap.generate_all_substitutional_impurity( 'V') # generate_all_vacancies generate_all_substitutional_impurity generate_all_interstitial


for a, b in zip(all_configs, all_labels):
    print(a.atomCountByType, b)
sadfsafd



path = '/home/akaris/Documents/code/Physics/VASP/v6.1/files/POSCAR/Cristals/NiOOH/*OH surface for pure NiOOH'
ap = CrystalDefectGenerator(file_location=path+'/SUPERCELL')
ap.readSIFile()
ap.is_surface = True
print( ap.latticeType )
#ap.introduce_vacancy(atom_index=10)
res,_ = ap.generate_all_vacancies()

ap.exportAsPOSCAR(path+'/POSCAR_d1')
for i, n in enumerate(res):
    n.exportAsPOSCAR(path+f'/POSCAR_d{i}')
'''



try:
    import copy 
    import pickle
    import numpy as np
    import functools, time
    from scipy.spatial import Voronoi
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing copy: {str(e)}\n")
    del sys

from .BasePartition import BasePartition

def timeit_decorator(func):
    """Decorator to measure the execution time of a function."""
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()  # Record the start time
        result = func(*args, **kwargs)  # Execute the function
        end_time = time.time()  # Record the end time
        execution_time = end_time - start_time  # Calculate the execution time
        print(f"Execution time of {func.__name__}: {execution_time} seconds")
        return result
    return wrapper

class Extract_builder(BasePartition):
    def __init__(self, file_location: str = None, name: str = None, *args, **kwargs):
        """
        Initializes the CrystalDefectBuilder object. If a Periodic_Object is provided, its attributes are copied.

        Parameters:
            file_location (str): Location of the file.
            name (str): Name of the object.
            Periodic_Object (object): An object containing periodic attributes.
            **kwargs: Additional keyword arguments.
        """
        super().__init__(*args, **kwargs)

    @staticmethod
    def get_atom_ids(AtomPositionManager, selection_list):
        """
        Retrieves atom IDs based on the selection list which can contain atom indices or labels.

        Parameters:
            AtomPositionManager (object): An object managing atom positions and labels.
            selection_list (list): A list of atom indices or labels.

        Returns:
            list: A list of unique atom IDs.
        """
        is_number = lambda s: s.replace('.', '', 1).isdigit() or s.replace('-', '', 1).replace('.', '', 1).isdigit()

        atom_ids = set()
        for item in selection_list:
            if isinstance(item, int) or is_number(item):
                atom_ids.add( int(item) )
            elif isinstance(item, str):
                if item.upper() in ['H2O', 'water']:
                    for i, label in enumerate(AtomPositionManager.atomLabelsList):
                        if label == 'O':
                            pos = AtomPositionManager.atomPositions[i]
                            closest_neighbors = AtomPositionManager.find_n_closest_neighbors(pos, 4)
                            if AtomPositionManager.atomLabelsList[closest_neighbors[1][1]] == AtomPositionManager.atomLabelsList[closest_neighbors[1][2]] == 'H' and closest_neighbors[0][2] < 1.2 and closest_neighbors[0][3] > 1.2: 
                                #if AtomPositionManager.atomPositions[i][2] > 26:
                                atom_ids.add( tuple(closest_neighbors[1][:3]) )

                if item.upper() in ['OOH']:
                    for i, label in enumerate(AtomPositionManager.atomLabelsList):
                        if label == 'O':
                            pos = AtomPositionManager.atomPositions[i]
                            closest_neighbors = AtomPositionManager.find_n_closest_neighbors(pos, 4)
                            if AtomPositionManager.atomLabelsList[closest_neighbors[1][1]] == 'H' and AtomPositionManager.atomLabelsList[closest_neighbors[1][2]] == 'O' and closest_neighbors[0][2] < 2.2 and closest_neighbors[0][3] > 2.2: 
                                #if AtomPositionManager.atomPositions[i][2] > 26:
                                atom_ids.add( tuple(closest_neighbors[1][:3]) )

                elif '-' in item:
                    a1, a2 = item.split('-')
                    for i, label in enumerate(AtomPositionManager.atomLabelsList):
                        if label == a1:
                            
                            pos = AtomPositionManager.atomPositions[i]
                            if AtomPositionManager.atomLabelsList[AtomPositionManager.find_n_closest_neighbors(pos, 2)[1][1]] == a2 or \
                                AtomPositionManager.atomLabelsList[AtomPositionManager.find_n_closest_neighbors(pos, 3)[1][2]] == a2 :
                                atom_ids.add( int(i) )

                else:
                    atom_ids.update([i for i, label in enumerate(AtomPositionManager.atomLabelsList) if label == item])
        
        return list(atom_ids) 

    def generate_extract(self, values, verbose:bool=False, containers:list=None):
        """
        Generates the configurational space by creating various combinations of defects in the crystal structure.

        Parameters:
            values (dict): A dictionary containing parameters for generating the configurational space.

        Returns:
            list: A list of containers with updated configurations.
        """
        containers = containers if isinstance(containers, list) else self.containers

        if values.get('last', False): containers = [containers[-1]]

        group = values['g']

        for container_index, container in enumerate(containers):
            
            atom_groups_ID = self.get_atom_ids(container.AtomPositionManager, group) 
            return atom_groups_ID

            if verbose: print(f' - Processing : {len(atom_groups_ID)} ')    
            print(group, atom_groups_ID)
            
            charge = container.AtomPositionManager.charge[atom_groups_ID][:,-1]
            magnetization = container.AtomPositionManager.magnetization[atom_groups_ID][:,-1]
            dist_matrix = np.array([ [container.AtomPositionManager.distance(container.AtomPositionManager.atomPositions[n1], 
                container.AtomPositionManager.atomPositions[n2]) if n1 > n2 else 0  for i2, n2 in enumerate(atom_groups_ID) ] for i1, n1 in enumerate(atom_groups_ID) ], dtype=np.float64 )
            labels = container.AtomPositionManager.atomLabelsList[atom_groups_ID]
            E = container.AtomPositionManager.E

            dictionary = { 
                'charge'        :     charge,
                'magnetization' :     magnetization,
                'dist_matrix'   :     dist_matrix,
                'labels'        :     labels,
                'E'             :     E,
                }

            print(dictionary)

            file_path = 'data'
            with open(file_path, 'wb') as file:
                pickle.dump(dictionary, file)
            
            print(f"Dictionary saved to {file_path}")















try:
    from ..master.FileManager import FileManager
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing FileManager: {str(e)}\n")
    del sys

try:
    from ..master.AtomicProperties import AtomicProperties
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing AtomicProperties: {str(e)}\n")
    del sys
    
try:
    import numpy as np
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing numpy: {str(e)}\n")
    del sys

class KPointsManager(FileManager, AtomicProperties):
    """
    A class to manage k-point data for use in electronic structure calculations (e.g., VASP).

    The KPointsManager class provides utilities for parsing, managing, and exporting
    k-point grids, band paths, and special points. It supports different k-point generation
    schemes, including regular grids (e.g., Monkhorst-Pack), automatic meshes, explicit k-points,
    and band structure paths.

    Attributes:
    ----------
    _num_kpoints : int
        Number of k-points in the grid or path.
    _coordinate_system : str
        The coordinate system used (e.g., Cartesian, Reciprocal).
    _kpoints : list
        List of k-points with optional weights.
    _mesh_type : str
        Type of k-point mesh ('Regular', 'Automatic', 'Explicit', etc.).
    _centering : str
        Centering of the mesh ('Gamma', 'Monkhorst-Pack').
    _subdivisions : list
        Subdivisions along the reciprocal lattice vectors.
    _shift : list
        Shift of the grid from the origin.
    _points_per_line : int
        Number of points per line in band structure calculations.
    _numberOfTetrahedra : int
        Number of tetrahedra in the grid (for tetrahedron integration).
    _volumeWeight : list
        Volume weights associated with the tetrahedra.
    _paths : list
        High-symmetry paths for band structure calculations.
    _special_lattice_points : dict
        Dictionary of special lattice points in reciprocal space.
    """
    def __init__(self, file_location:str=None, name:str=None, **kwargs):
        """
        Initialize the KPointsManager class.

        Parameters:
        ----------
        file_location : str, optional
            Path to the KPOINTS file to read.
        name : str, optional
            Name identifier for the k-point set.
        """
        FileManager.__init__(self, name=name, file_location=file_location)
        AtomicProperties.__init__(self)

        self._num_kpoints = None
        self._coordinate_system = None
        self._kpoints = []
        self._mesh_type = None
        self._centering = None
        self._subdivisions = None
        self._shift = None
        self._points_per_line = None

        self._numberOfTetrahedra = None
        self._volumeWeight = None
        self._KPOINTmesh = None

        self._paths = []  # Stores band path as segments of pairs
        self._special_lattice_points = {}  # Predefined special points

    @property
    def band_path(self):
        """Getter for band path."""
        return self._paths

    @band_path.setter
    def band_path(self, path):
        """
        Setter for band path. Converts a list of points into path segments.

        Parameters:
        ----------
        path : list
            List of point labels corresponding to a predefined set of special points.
        """
        if not isinstance(path, list):
            raise ValueError("Band path must be a list of point labels.")
        if len(path) < 2:
            raise ValueError("Band path must contain at least two points.")

        self._paths = []
        for i in range(len(path) - 1):  # Crear segmentos consecutivos
            start_label, end_label = path[i], path[i + 1]
            if start_label not in self._special_lattice_points or end_label not in self._special_lattice_points:
                raise ValueError(f"Labels {start_label} or {end_label} not found in special points.")
            start_point = self._special_lattice_points[start_label] + [start_label]
            end_point = self._special_lattice_points[end_label] + [end_label]
            self._paths.append([start_point, end_point])

    @property
    def special_points(self):
        """Getter for special points."""
        return self._special_lattice_points

    @special_points.setter
    def special_points(self, points_dict):
        """
        Setter for special lattice points.

        Parameters:
        ----------
        points_dict : dict
            A dictionary of special lattice points with labels as keys
            and reciprocal coordinates as values.
        """
        if not isinstance(points_dict, dict):
            raise ValueError("Special points must be provided as a dictionary.")
        self._special_lattice_points = points_dict


    def readKPOINTS(self, file_location:str=None):
        """
        Reads a KPOINTS file and parses its content.

        Parameters:
        ----------
        file_location : str, optional
            Path to the KPOINTS file. If not provided, the class instance's `file_location` is used.

        Notes:
        -----
        - The KPOINTS file determines the sampling of reciprocal space for electronic
          structure calculations. Depending on the mesh type, different parsing logic
          is applied.
        """
        file_location = file_location if type(file_location) == str else self.file_location

        lines = [n for n in self.read_file(file_location) ]

        self.comment = lines[0].strip()
        self.num_kpoints = int(float(lines[1].strip().split()[0]))

        mesh_type_flag = lines[2].strip().split()[0].lower()[0]    

        if self.num_kpoints == 0:
            if mesh_type_flag in ['g', 'm']:
                self.mesh_type = 'Regular'
                self.centering = 'Gamma' if mesh_type_flag == 'g' else 'Monkhorst-Pack'
                self.subdivisions = [int(float(n)) for n in lines[3].strip().split()[:3] ] #list(map(float, lines[3].strip().split()[:3]) )
                if len(lines) > 4:
                    try: self.shift = list(map(float, lines[4].strip().split()[:3] ))
                    except: pass

            elif mesh_type_flag in ['a']:
                self.mesh_type = 'Automatic'
                self.length = float(lines[3].strip())

            else:
                self.mesh_type = 'Generalized'
                self.coordinate_system = 'Cartesian' if mesh_type_flag in ['c', 'k'] else 'Fractional' if mesh_type_flag == 'r' else None
                self.KPOINTmesh = []
                for i, line in enumerate(lines[3:]):  
                    line_split = line.strip().split()
                    if len(line_split) > 2:
                        self.KPOINTmesh.append(line_split[:3]) 

        elif mesh_type_flag == 'l':
            self.mesh_type = 'Line Mode'
            self.points_per_line = self.num_kpoints
            self.coordinate_system = lines[3].strip()
            path = []
            for line in lines[4:]:
                if line.strip():  # Skip empty lines
                    coords, label = line.strip().rsplit(maxsplit=1)
                    kx, ky, kz = map(float, coords.split())
                    path.append({'kx': kx, 'ky': ky, 'kz': kz, 'label': label})
                else:
                    if path:
                        self.paths.append(path)
                        path = []
            if path:
                self.paths.append(path)  # Append remaining path if any

        else:
            self.mesh_type = 'Explicit'

            if mesh_type_flag == 'c' or mesh_type_flag == 'k':
                self.coordinate_system = 'Cartesian'
            else:        
                self.coordinate_system = 'Fractional'

            for i, line in enumerate(lines[3:]):  
                if line.strip().split()[0][0].lower() == 't': # Tetrahedron case
                    self.numberOfTetrahedra = lines[i+1].strip().split(' ')[0]
                    self.volumeWeight = lines[i+1].strip().split(' ')[1:]
                    break
                else:
                    kx, ky, kz, weight = map(float, line.split()[:4])
                    self.kpoints.append({'kx': kx, 'ky': ky, 'kz': kz, 'weight': weight})

    def exportAsKPOINTS(self, file_location: str = None):
        """
        Exports the k-point information to a VASP-compatible `KPOINTS` file.

        Parameters:
        ----------
        file_location : str, optional
            The path to save the exported `KPOINTS` file. If not provided,
            the instance's `file_location` will be used.

        Notes:
        -----
        - The format of the exported file depends on the `mesh_type`:
            - 'Regular': Gamma or Monkhorst-Pack mesh.
            - 'Automatic': A single number specifying the mesh density.
            - 'Line Mode': High-symmetry paths for band structure calculations.
            - 'Explicit': A manually specified list of k-points.
        - Each path in 'Line Mode' includes comments with the corresponding
          special point labels.
        """
        # Determine the output file location, using the instance's `file_location` as fallback
        file_location = file_location or self.file_location
        
        # Initialize the lines list with the comment and number of k-points
        lines = [self.comment, str(self.num_kpoints)]

        # Handle different k-point mesh types
        if self.mesh_type == 'Regular':
            # Regular mesh (Gamma or Monkhorst-Pack)
            lines.append(self.centering)
            lines.append(" ".join(map(str, self.subdivisions)))
            if self.shift:  # Include shift if defined
                lines.append(" ".join(map(str, self.shift)))

        elif self.mesh_type == 'Automatic':
            # Automatic mesh with density specified by `self.length`
            lines.extend(['Auto', str(self.length)])

        elif self.mesh_type == 'Line Mode':
            # Band structure paths along high-symmetry points
            lines.extend(['Line Mode', self._coordinate_system])
            # Construct path lines with special point labels
            path_lines = [
                f"{start[0]} {start[1]} {start[2]} ! {start[3]}\n{end[0]} {end[1]} {end[2]} ! {end[3]}\n"
                for start, end in self._paths
            ]
            lines.extend(path_lines)

        elif self.mesh_type == 'Explicit':
            # Explicitly specified k-points with weights
            lines.append(self.coordinate_system)
            lines.extend(
                f"{kpoint['kx']} {kpoint['ky']} {kpoint['kz']} {kpoint['weight']}"
                for kpoint in self.kpoints
            )

        # Write the collected lines to the output file
        with open(file_location, 'w') as f:
            f.write("\n".join(lines) + "\n")  # Ensure proper newline formatting


    def set_band_path(self, special_points: dict, segments: list, divisions:int=30):
        """
        Configure the band path using special points and connecting segments.

        Parameters:
        ----------
        special_points : dict
            Dictionary of special points in reciprocal space.
        segments : list
            List of tuples representing the path segments [(label1, label2), ...].
        divisions : int
            Number of points per segment.
        """
        self.comment = 'KPOINTS for band structure'
        self.num_kpoints = divisions
        self.mesh_type = 'Line Mode'
        self.coordinate_system = 'Reciprocal'

        self._special_lattice_points = special_points  # Store the special points
        self._paths = []  # Clear existing paths

        for start_label, end_label in segments:
            if start_label not in special_points or end_label not in special_points:
                raise ValueError(f"Labels {start_label} or {end_label} not found in special points.")
            start_point = special_points[start_label] + [start_label]
            end_point = special_points[end_label] + [end_label]
            self._paths.append([start_point, end_point])

        return True

    def summary(self, ) -> str:
        """
        Provides a summary of the parsed k-point information.

        Returns:
        -------
        str
            A formatted string describing the k-point setup.
        """
        text_str = ''
        # Display parsed information based on the mesh type
        if self.mesh_type == 'Explicit':
            text_str +=f"Mesh Type: {self.mesh_type}\n"
            text_str += f"Comment: {self.comment}\n"
            text_str += f"Number of k-points: {self.num_kpoints}\n"
            text_str += f"Coordinate System: {self.coordinate_system}\n"
            text_str += f"K-points: {self.kpoints}\n"
            text_str += f"NumberOfTetrahedra: {self.numberOfTetrahedra}\n"
            text_str += f"VolumeWeight: {self.volumeWeight}\n"

        elif self._mesh_type == 'Line Mode':
            text_str += f"Mesh Type: {self._mesh_type}\n"
            text_str += f"Comment: {self.comment}\n"
            text_str += f"Points per Line: {self._points_per_line}\n"
            text_str += f"Coordinate System: {self._coordinate_system}\n"
            text_str += f"Paths:\n"
            for i, path in enumerate(self._paths):
                k0, k1 = path[0], path[1]
                text_str += f' ({i}) {k0[0]:<4.2f},{k0[1]:<4.2f},{k0[2]:<4.2f} ({k0[3]}) >> {k1[0]:<4.2f},{k1[1]:<4.2f},{k1[2]:<4.2f} ({k1[3]})\n'
         
        elif self.mesh_type == 'Automatic':
            text_str += f"Mesh Type: {self.mesh_type}\n"
            text_str += f"Comment: {self.comment}\n"
            text_str += f"Length: {self.length}\n"

        elif self.mesh_type == 'Generalized':
            text_str += f"Mesh Type: {self.mesh_type}\n"
            text_str += f"Comment: {self.comment}\n"
            text_str += f"Length: {self.coordinate_system}\n"
            text_str += f"Length: {self.KPOINTmesh}\n"
        else:
            text_str += f"Mesh Type: {self.mesh_type}\n"
            text_str += f"Comment: {self.comment}\n"
            text_str += f"Centering: {self.centering}\n"
            text_str += f"Subdivisions: {self.subdivisions}\n"
            if self.shift:
                text_str += f"Shift: {self.shift}\n"
        return text_str 

if __name__ == "__main__":
    # Example: Hexagonal cell
    cell_hex = np.array([[1.0, 0.0, 0.0],               # Vector a
                         [-0.5, np.sqrt(3) / 2, 0.0],   # Vector b
                         [0.0, 0.0, 2.0]])              # Vector c

    # Identify the hexagonal lattice
    try:
        from ..miscellaneous.get_bravais_lattice import identify_lattice, ibz_points, sc_special_points, special_segments
        lattice, operation = identify_lattice(cell_hex, eps=1e-4, pbc=[True, True, True])
        print(ibz_points[lattice.name], sc_special_points[lattice.name], special_segments[lattice.name])
        print(f"The identified lattice type is: {lattice.name}")
        print (  )
        KPOITNS = KPointsManager()
        KPOITNS.set_band_path( sc_special_points[lattice.name], special_segments[lattice.name] )
        #KPOITNS.exportAsKPOINTS('KPOINTS')

    except RuntimeError as e:
        print(f"Could not identify the lattice: {e}")

'''
kpoints_manager = KPointsManager('/home/akaris/Documents/code/Physics/VASP/v6.1/files/KPOINTS/example/KPOINTband')
kpoints_manager = KPointsManager('/home/akaris/Documents/code/Physics/VASP/v6.1/files/KPOINTS/example/KPOINTSgeneralized')
kpoints_manager = KPointsManager('/home/akaris/Documents/code/Physics/VASP/v6.1/files/KPOINTS/example/KPOINTregular')
kpoints_manager = KPointsManager('/home/akaris/Documents/code/Physics/VASP/v6.1/files/KPOINTS/example/KPOINTSexplicitTetrahedron')
kpoints_manager = KPointsManager('/home/akaris/Documents/code/Physics/VASP/v6.1/files/KPOINTS/example/KPOINTSexplicit')
kpoints_manager.readKPOINTS()
print(kpoints_manager.summary())
kpoints_manager.exportKPOINTS('/home/akaris/Documents/code/Physics/VASP/v6.1/files/KPOINTS/example/KPOINTexportation')
'''



try:
    import numpy as np
    from tqdm import tqdm
    import copy
    from typing import List, Dict, Tuple

    from .BasePartition import BasePartition
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing numpy: {str(e)}\n")
    del sys

class PositionEditor_builder(BasePartition):
    """
    A class for managing and editing atomic positions in various ways, inheriting from PartitionManager.

    Methods
    -------
    handle_rattle(values, file_location=None)
        Applies a random displacement to atomic positions.

    handle_compress(container, values, container_index, file_location=None)
        Compresses a given container along a defined direction.

    handle_widening(values, file_location=None)
        Stacks atoms from different containers to widen atomic layers.

    handle_interpolation(values, file_location=None)
        Interpolates between atomic positions using splines.

    handle_exfoliation(values, file_location=None)
        Separates atomic layers to simulate exfoliation.
    """
    def __init__(self, file_location: str = None, name: str = None, *args, **kwargs):
        """
        Initialize the PositionEditorBuilder instance.

        Parameters
        ----------
        file_location : str, optional
            The file location for the atomic data.
        name : str, optional
            The name for this instance.
        """
        super().__init__(*args, **kwargs)

    def handle_rattle(self, values, file_location=None):
        """
        Applies random displacements to atom positions within containers to simulate atomic 'rattling'.
        
        Parameters
        ----------
        values : dict
            Contains parameters for the rattling process (e.g., standard deviation 'std', number of iterations 'N').
        file_location : str, optional
            Directory to store rattled configurations.

        Returns
        -------
        list
            A list of containers with rattled atom positions.
        """
        containers = []

        # Iterate through each container and apply random rattling 'N' times.
        for container_index, container in enumerate(self.containers):
            for n in range(values['N']):
                # Create a copy of the container and apply the random displacement.
                container_copy = self.copy_and_update_container(container, f'/rattle/{container_index}_{n}', file_location)
                container_copy.AtomPositionManager.rattle(stdev=values['std'], seed=n)

                containers.append(container_copy)
        
        # Update the current containers with the rattled containers.
        self.containers = containers

        return containers

    def handle_compress(self, values, file_location=None):
        """
        Compresses the atom positions within a container by applying different compression factors.
        
        Parameters
        ----------
        container : object
            The container whose atom positions will be compressed.
        values : dict
            Contains parameters like 'compress_min', 'compress_max', and 'N' for the compression range.
        container_index : int
            Index of the container.
        file_location : str, optional
            Directory to store compressed configurations.

        Returns
        -------
        list
            A list of containers with compressed atom positions.
        """
        sub_directories, containers = [], []

        # Create a vector of compression factors.
        compress_vector = self.interpolate_vectors(values['compress_min'], values['compress_max'], values['N'])

        # Iterate through each container and compress.
        for container_index, container in enumerate(tqdm(self.containers, desc="Processing Containers")):

            for v_i, v in enumerate(compress_vector):
                # Create a copy of the container and apply the compression factor.
                container_copy = self.copy_and_update_container(container, f'/compress/{v_i}', file_location)
                container_copy.AtomPositionManager.compress(compress_factor=v, verbose=False)
                    
                sub_directories.append(f'/{v_i}')
                containers.append(container_copy)

        return containers

    def handle_widening(self, values, file_location=None):
        """
        Stacks multiple atomic configurations to create wider layers of atoms.
        
        Parameters
        ----------
        values : list of dict
            Each dictionary contains indices and stacking direction information.
        file_location : str, optional
            Directory to store widened configurations.

        Returns
        -------
        list
            A list of containers with widened atomic layers.
        """
        sub_directories, containers = [], []

        # Iterate over the provided values to stack atomic configurations.
        for v_i, v in enumerate(values):
            container_init = self.containers[v['init_index']]
            container_mid = self.containers[v['mid_index']]
            container_end = self.containers[v['end_index']]
            
            # Create a copy of the initial container.
            container_copy = self.copy_and_update_container(container_init, f'/widening/{v_i}', file_location)

            # Stack the middle container 'N' times, then add the end container.
            for n in range(v['N']):
                container_copy.AtomPositionManager.stack(AtomPositionManager=container_mid.AtomPositionManager, direction=v['direction'])
            container_copy.AtomPositionManager.stack(AtomPositionManager=container_end.AtomPositionManager, direction=v['direction'])

            sub_directories.append(f'/{v_i}')
            containers.append(container_copy)
    
        return containers

    def handle_interpolation(self, values, file_location=None):
        """
        Interpolates atom positions between given configurations using spline interpolation.
        
        Parameters
        ----------
        values : dict
            Contains parameters for interpolation like 'images' and 'degree'.
        file_location : str, optional
            Directory to store interpolated configurations.

        Returns
        -------
        list
            A list of containers with interpolated atom positions.
        """
        interpolation_data = np.zeros((self.containers[0].AtomPositionManager.atomCount, 3, len(self.containers)), dtype=np.float64)

        # Collect atomic positions from each container.
        for container_index, container in enumerate(self.containers):
            if values.get('first_neighbor', False):
                container.AtomPositionManager.wrap()
            interpolation_data[:, :, container_index] = container.AtomPositionManager.atomPositions_fractional

        # Adjust positions for periodic boundaries if specified.
        if values.get('first_neighbor', False):
            diff = np.diff(interpolation_data, axis=2)
            interpolation_data[:, :, 1:][diff > 0.5] -= 1
            interpolation_data[:, :, 1:][diff < -0.5] += 1

        # Perform spline interpolation on the collected data. 
        new_interpolated_data = self.interpolate_with_splines(interpolation_data, M=values['images'], degree=values['degree'])

        containers = [self.copy_and_update_container(self.containers[0], f'/interpolation/init', file_location)]
        for container_index, container in enumerate(self.containers[1:]):
            for n in range(values['images'] + 1):
                container_copy = self.copy_and_update_container(container, f'/interpolation/{container_index}_{n}', file_location)
                container_copy.AtomPositionManager.set_atomPositions_fractional(new_interpolated_data[:, :, container_index * (values['images'] + 1) + n + 1])
                if values.get('first_neighbor', False):
                    container_copy.AtomPositionManager.wrap()
                containers.append(container_copy)

        apm_cr = containers[0].AtomPositionManager.covalent_radii
        pair_min_dist = {(ua1, ua2): (apm_cr[ua1]+apm_cr[ua2])*.9  for ua1 in self.uniqueAtomLabels for ua2 in self.uniqueAtomLabels }

        self._physics_relax_images(
            containers,
            first_neighbor=values.get('first_neighbor', False),
            # per-pair hard minima (optional)
            pair_min_dist=values.get('pair_min_dist', None ),
            # repulsive wall scale r0_ij = s_rep * (r_i + r_j) unless pair_min_dist overrides it
            s_rep=values.get('s_rep', 0.85),
            k_rep=values.get('k_rep', 100.0),     # eV/Å^2 (quadratic wall strength)
            k_tether=values.get('k_tether', 2.0), # eV/Å^2 (how tightly to follow the spline)
            max_steps=values.get('relax_steps', 60),
            step0=values.get('relax_step0', 0.10),# initial step in Å
            gtol=values.get('relax_gtol', 1e-2),  # max |grad| stop
            ls_shrink=values.get('ls_shrink', 0.5),
            ls_expand=values.get('ls_expand', 1.2)
        )

        self.set_container(containers)
        return containers

    def handle_exfoliation(self, values, file_location=None):
        """
        Simulates the exfoliation of atomic layers by separating them along a specified axis.
        
        Parameters
        ----------
        values : dict
            Contains parameters such as 'direction' and 'threshold' to determine layers.
        file_location : str, optional
            Directory to store exfoliated configurations.

        Returns
        -------
        list
            A list of containers with exfoliated atomic layers.
        """
        containers = []
        separation_distance = values.get('separation_distance', 20.0)   # Default separation distance between layers.
        layer_direction = values.get('direction', 'y') 
        layer_threshold = values.get('threshold', 2.0) 

        # Iterate through each container and separate layers.
        for container_index, container in enumerate(tqdm( copy.deepcopy(self.containers), desc="Processing Containers")):
            # Wrap atom positions within simulation boundaries to ensure all atoms are inside the box.
            container.AtomPositionManager.wrap()

            # Get indices of atoms in each layer based on the specified direction and threshold.
            layers_index_list = container.AtomPositionManager.get_layers(direction=layer_direction, 
                                                                         threshold=layer_threshold)
            
            # Create a deep copy of the original container to add solvent in the slab form.
            container_original = copy.deepcopy(container)
            
            container_original_water = self.handleCLUSTER(    values = {'ADD_SOLVENT':{
                                            'density': 1.01,
                                            'solvent': ['H2O'],
                                            'slab': True,
                                            'shape': None,
                                            'size': None,
                                            'vacuum_tolerance': 0,
                                            'colition_tolerance': 1.75,
                                            'translation': None,
                                            'wrap': True,
                                            'seed':0,
                                            'max_iteration':100000,
                                            'verbose':True
                                        }}, 
                                    containers=container_original)

            # Add the modified container with water to the list of results.
            containers += container_original_water

            # Iterate through possible separation points and generate configurations for exfoliated layers.
            for cut_index in range(1, len(layers_index_list)):
                # Create a deep copy of the container for each new configuration.
                container_copy = copy.deepcopy(container)

                # Move atoms in different directions to simulate the exfoliation process.
                for i, layer_indices in enumerate(layers_index_list):
                    if i > cut_index:
                        container_copy.AtomPositionManager.move_atom(layer_indices, [0, separation_distance / 2, 0])
                    #else:
                    #    container_copy.AtomPositionManager.move_atom(layer_indices, [0, -separation_distance / 2, 0])

                # Determine the vacuum box dimensions (used if further modification is needed).
                vacuum_box, vacuum_start = container.AtomPositionManager.get_vacuum_box(tolerance=0)
                
                # Add solvent to the exfoliated container using more customizable parameters.
                container_copy_water = self.handleCLUSTER(values={'ADD_SOLVENT': {
                                        'density': 1.0,
                                        'solvent': ['H2O'],
                                        'slab': False,
                                        'shape': 'CELL',
                                        'size': None,
                                        'vacuum_tolerance': 0,
                                        'colition_tolerance': 1.75,
                                        'molecules_number':self.molecules_number,
                                        'translation': None,
                                        'wrap': True,
                                        'seed':0,
                                        'max_iteration':100000,
                                        'verbose':True
                                    }}, 

                                containers=[container_copy])

                # Attempt to add the modified container with water to the list of results.
                try:
                    containers += container_copy_water
                except:
                    print('Warning: Could not add solvent to the container. Consider trying a lower density.')
                        
        return containers


    def _physics_relax_images(
        self,
        containers: List[object],
        *,
        first_neighbor: bool,
        pair_min_dist: Dict[Tuple[str, str], float],
        s_rep: float,
        k_rep: float,
        k_tether: float,
        max_steps: int,
        step0: float,
        gtol: float,
        ls_shrink: float,
        ls_expand: float,
    ) -> None:
        """Relax each interpolated image with a cheap repulsive + tether potential."""
        if not containers:
            return

        # Precompute covalent radii per-atom from the first container
        labels0 = containers[0].AtomPositionManager.atomLabelsList
        cov_rad = np.array([containers[0].AtomPositionManager.covalent_radii[el] for el in labels0], float)
        N = len(labels0)

        # Build reference fractional positions for each image (the current positions)
        refs_frac = [c.AtomPositionManager.atomPositions_fractional.copy() for c in containers]

        for idx, cont in enumerate(containers):
            # Relax every image except maybe the very first/last if you want to pin them
            # (comment out these lines if you want to relax all)
            # if idx == 0 or idx == len(containers)-1:
            #     continue
            x_ref_f = refs_frac[idx]
            self._physics_relax_one(
                cont, x_ref_f, labels0, cov_rad,
                first_neighbor=first_neighbor,
                pair_min_dist=pair_min_dist,
                s_rep=s_rep, k_rep=k_rep,
                k_tether=k_tether,
                max_steps=max_steps, step0=step0,
                gtol=gtol, ls_shrink=ls_shrink, ls_expand=ls_expand
            )


    def _physics_relax_one(
        self, container, x_ref_frac, labels, cov_rad,
        *, first_neighbor: bool,
        pair_min_dist: Dict[Tuple[str, str], float],
        s_rep: float, k_rep: float, k_tether: float,
        max_steps: int, step0: float, gtol: float,
        ls_shrink: float, ls_expand: float
    ) -> None:
        """Minimize E = Σ_{i<j} 0.5*k_rep*max(0, r0_ij - r_ij)^2 + 0.5*k_tether*Σ_i ||x_i - x_i^ref||^2."""
        # Work in Cartesian for gradients; keep fractional for wrapping/setting
        x_f = container.AtomPositionManager.atomPositions_fractional.copy()
        x = _frac_to_cartesian(container, x_f)           # (N,3)
        x_ref = _frac_to_cartesian(container, x_ref_frac)# (N,3)
        N = x.shape[0]

        def r0_ij(i, j):
            # explicit per-pair override in Å takes precedence
            if pair_min_dist:
                key = (labels[i], labels[j])
                keyr = (labels[j], labels[i])
                if key in pair_min_dist: return pair_min_dist[key]
                if keyr in pair_min_dist: return pair_min_dist[keyr]
            return s_rep * (cov_rad[i] + cov_rad[j])

        # simple backtracking line search gradient descent
        step = float(step0)

        for it in range(max_steps):
            E, g = _energy_and_grad_rep_tether(
                container, x, x_ref, labels, cov_rad,
                r0_ij_fn=r0_ij,
                k_rep=k_rep, k_tether=k_tether,
                first_neighbor=first_neighbor
            )
            gmax = float(np.abs(g).max())
            if gmax < gtol:
                break

            # trial step with backtracking
            desc_dir = -g
            ok = False
            for _ in range(20):
                x_trial = x + step * desc_dir
                # project to fractional, wrap if needed, then back to Cartesian (keeps PBC sane)
                x_trial_f = _cartesian_to_frac(container, x_trial)
                if first_neighbor:
                    x_trial_f = np.mod(x_trial_f, 1.0)
                x_trial = _frac_to_cartesian(container, x_trial_f)

                E_trial, _ = _energy_and_grad_rep_tether(
                    container, x_trial, x_ref, labels, cov_rad,
                    r0_ij_fn=r0_ij,
                    k_rep=k_rep, k_tether=k_tether,
                    first_neighbor=first_neighbor
                )
                if E_trial <= E:
                    x = x_trial
                    step *= ls_expand
                    ok = True
                    break
                step *= ls_shrink

            if not ok:
                # step became too small; stop
                break

        # Commit positions
        x_f_new = _cartesian_to_frac(container, x)
        container.AtomPositionManager.set_atomPositions_fractional(x_f_new)
        if first_neighbor:
            container.AtomPositionManager.wrap()


def _energy_and_grad_rep_tether(
    container, x, x_ref, labels, cov_rad,
    *, r0_ij_fn, k_rep: float, k_tether: float, first_neighbor: bool
):
    """Return (E, grad) for repulsive wall + path tether."""
    N = x.shape[0]
    g = np.zeros_like(x)
    E = 0.0

    # Pair repulsion: quadratic wall when r < r0_ij; 0 otherwise
    for i in range(N - 1):
        for j in range(i + 1, N):
            # minimum-image vector in fractional → Cartesian
            vij_f = _cartesian_to_frac(container, x[j:j+1] - x[i:i+1])[0]
            if first_neighbor:
                vij_f = _min_image_fractional(vij_f)
            rij_vec = _frac_to_cartesian(container, vij_f[None, :])[0]
            rij = float(np.linalg.norm(rij_vec))
            if rij < 1e-12:
                # avoid singular; small random axis
                n = np.array([1.0, 0.0, 0.0], float); rij = 1e-12
            else:
                n = rij_vec / rij

            r0 = float(r0_ij_fn(i, j))
            if rij < r0:
                overlap = (r0 - rij)
                E += 0.5 * k_rep * overlap * overlap
                dE_drij = -k_rep * overlap              # d/d r
                # grad contribution: dE/dx_i = dE/dr * dr/dx_i = dE/dr * (-n)
                gi = dE_drij * (-n)
                gj = dE_drij * (+n)
                g[i] += gi
                g[j] += gj

    # Path tether: harmonic to the interpolated geometry
    dx = x - x_ref
    E += 0.5 * k_tether * float(np.sum(dx * dx))
    g += k_tether * dx

    return E, g




def _min_image_fractional(dfrac: np.ndarray) -> np.ndarray:
    """Map fractional deltas to [-0.5, 0.5] per component."""
    out = dfrac.copy()
    out[out > 0.5] -= 1.0
    out[out < -0.5] += 1.0
    return out


def _build_identity_mapping(
    self,
    A,
    B,
    radii: np.ndarray,
    *,
    id_scale: float,
    first_neighbor: bool
) -> np.ndarray:
    """
    Greedy, radii-aware nearest-neighbour matching under PBC if requested.
    If no candidate satisfies the radii cutoff, assigns the nearest available neighbour.

    Returns
    -------
    np.ndarray
        mapping[i] = j  (indices into B)
    """
    N = len(radii)
    used = np.zeros(N, dtype=bool)
    mapping = -np.ones(N, dtype=int)

    # Precompute a distance matrix using your self.distance helper.
    # Adjust the call signature if your API differs.
    D = np.empty((N, N), dtype=float)
    for i in range(N):
        for j in range(N):
            # Expecting self.distance to respect PBC when first_neighbor is True.
            # If it needs fractional coords instead, adapt this call accordingly.
            D[i, j] = A.AtomPositionManager.distance(
                A.AtomPositionManager.atomPositions[i], 
                B.AtomPositionManager.atomPositions[j]
            )

    # For each i, preferred js sorted by distance
    pref = np.argsort(D, axis=1)

    # First pass: enforce covalent-radii identity (tight) when possible.
    cutoff = id_scale * (radii[:, None] + radii[None, :])  # shape (N, N)
    for i in range(N):
        assigned = False
        for j in pref[i]:
            if used[j]:
                continue
            if D[i, j] <= cutoff[i, j]:
                mapping[i] = j
                used[j] = True
                assigned = True
                break
        if not assigned:
            # Defer to second pass.
            continue

    # Second pass: assign any remaining i to the nearest unused j.
    for i in range(N):
        if mapping[i] != -1:
            continue
        for j in pref[i]:
            if not used[j]:
                mapping[i] = j
                used[j] = True
                break

    # Safety: ensure total mapping.
    assert np.all(mapping >= 0), "Failed to build a complete atom mapping"
    return mapping


def _frac_to_cartesian(container, dfrac: np.ndarray) -> np.ndarray:
    """
    Convert fractional vectors to Cartesian using the container's lattice.
    Adapt the lattice accessor if your codebase uses a different attribute.
    """
    # Try common names for the 3x3 lattice matrix with rows as a,b,c in Cartesian Å.
    cell = getattr(container.AtomPositionManager, 'latticeVectors', None)
    if cell is None:
        cell = getattr(container.AtomPositionManager, 'cell', None)
    if cell is None:
        cell = getattr(container.AtomPositionManager, 'lattice_cartesian', None)
    if cell is None:
        raise RuntimeError("Cannot locate lattice matrix to convert fractional → Cartesian.")
    cell = np.asarray(cell, dtype=float)  # shape (3,3)
    return dfrac @ cell  # (N,3) x (3,3) → (N,3)


def _cartesian_to_frac(container, dcart: np.ndarray) -> np.ndarray:
    """Convert Cartesian vectors to fractional using the container's lattice."""
    cell = getattr(container, 'latticeVectors', None)
    if cell is None:
        cell = getattr(container.AtomPositionManager, 'latticeVectors', None)
    if cell is None:
        cell = getattr(container.AtomPositionManager, 'lattice_cartesian', None)
    if cell is None:
        raise RuntimeError("Cannot locate lattice matrix to convert Cartesian → fractional.")
    cell = np.asarray(cell, dtype=float)  # (3,3)
    # Solve cell^T * frac = cart^T  →  frac = cart @ inv(cell)
    inv_cell = np.linalg.inv(cell)
    return dcart @ inv_cell












try:
    from .SingleRunDFT import SingleRunDFT
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing SingleRunDFT: {str(e)}\n")
    del sys

class SingleRun(SingleRunDFT):  # Note: The class name is intentionally generalized and does not include "DFT".
    """
    A general class representing a single run simulation.

    Although this class inherits from SingleRunDFT, it is intended to serve as a more general 
    simulation container not restricted to Density Functional Theory (DFT). It manages various 
    simulation components such as atomic positions, input/output file handling, k-points, potentials,
    and additional handlers.

    Attributes
    ----------
    _loaded : dict
        Dictionary to track loaded modules or data.
    _AtomPositionManager : object
        Manager instance responsible for handling atomic positions.
    _Out_AtomPositionManager : object or None
        Manager for output atomic positions (if applicable).
    _InputFileManager : object or None
        Manager for input file operations.
    _KPointsManager : object or None
        Manager for handling k-point information in reciprocal space.
    _PotentialManager : object or None
        Manager for potential data used in the simulation.
    _BashScriptManager : object or None
        Manager for generating or executing bash scripts.
    _vdw_kernel_Handler : object or None
        Handler for van der Waals kernel processing.
    _OutFileManager : object or None
        Manager for output file operations.
    _WaveFileManager : object or None
        Manager for wavefunction file handling.
    _ChargeFileManager : object or None
        Manager for charge density file operations.
    """

    def __init__(self, file_location: str = None, name: str = None, **kwargs):
        """
        Initialize a SingleRun simulation instance.

        This constructor calls the parent initializer and sets up various internal managers 
        that are required to manage simulation inputs and outputs.

        Parameters
        ----------
        file_location : str, optional
            The file path used to locate initial configuration data.
        name : str, optional
            A unique identifier for the simulation instance.
        **kwargs
            Additional keyword arguments passed to the base class.
        """
        super().__init__(name=name, file_location=file_location)

        # Dictionary to store loaded modules or data for later use.
        self._loaded = {}
        self.file_location = file_location

        # Initialize the AtomPositionManager using its constructor; this handles atomic position data.
        self._AtomPositionManager = self.AtomPositionManager_constructor(self.file_location)
        
        # Initialize additional managers to None; these can be instantiated later as needed.
        self._Out_AtomPositionManager = None
        self._InputFileManager = None
        self._KPointsManager = None
        self._PotentialManager = None
        self._BashScriptManager = None
        self._vdw_kernel_Handler = None
        self._OutFileManager = None
        self._WaveFileManager = None
        self._ChargeFileManager = None

    @property
    def AtomPositionManager(self):
        """

        """
        if self._AtomPositionManager is None:
            self._AtomPositionManager = self.AtomPositionManager_constructor(self.file_location)

        return self._AtomPositionManager

    def read_structure(self, file_location, source, *args, **kwargs):
        """
        Read the structure from a file and update the AtomPositionManager.

        This method initializes the AtomPositionManager using the provided file location and 
        then reads the structural data using the specified source format. If any errors occur 
        during this process, they are caught and a message is printed.

        Parameters
        ----------
        file_location : str
            The path to the file containing the structure information.
        source : str
            A string identifier for the structure format or the reading method.
        *args
            Additional positional arguments.
        **kwargs
            Additional keyword arguments.
        """
        try:
            # Reinitialize the AtomPositionManager for the given file location.
            self.AtomPositionManager = self.AtomPositionManager_constructor(file_location)
            # Read the structure using the specified source.
            self.AtomPositionManager.read(source=source, file_location=file_location)
        except Exception as e:
            print(f"Failed to read {file_location} as {source}: {e}")

    def export_structure(self, file_location, source, *args, **kwargs):
        """
        Export the current structure to a file.

        This method reinitializes the AtomPositionManager for the given file location and then 
        exports the structure data in the specified format. Any encountered exceptions are caught 
        and reported via a printed message.

        Parameters
        ----------
        file_location : str
            The path to the file where the structure will be exported.
        source : str
            A string identifier for the structure format or the export method.
        *args
            Additional positional arguments.
        **kwargs
            Additional keyword arguments.
        """
        try:
            # Reinitialize the AtomPositionManager for the given file location.
            self.AtomPositionManager = self.AtomPositionManager_constructor(file_location)
            # Export the structure using the specified source.
            self.AtomPositionManager.export(source=source, file_location=file_location)
        except Exception as e:
            print(f"Failed to export {file_location} as {source}: {e}")




import numpy as np

def harmonic(kappa, xstar):
    """
    Returns vectorized umbrella potential
    U(x) = (kappa/2) * (x - xstar)^2
    """
    return lambda x: (kappa / 2) * np.square(x - xstar)

def linear(phi):
    """
    Returns vectorized umbrella potential
    U(x) = phi * x
    """
    return lambda x: phi * x

def harmonic_DD(kappa, xstar):
    """
    Returns vectorized umbrella potential
    U(x) = (1/2) * sum(kappa_i * (x_i - xstar_i)^2)
    kappa and xstar have shape (D,)
    x has shape (N, D)
    U(x) has shape (N,)
    """
    kappa = np.asarray(kappa).reshape(-1)
    xstar = np.asarray(xstar).reshape(-1)
    
    if kappa.shape != xstar.shape:
        raise ValueError(f"kappa and xstar must have the same shape, but got {kappa.shape} and {xstar.shape}")
    
    def potential(x):
        if x.ndim != 2 or x.shape[1] != kappa.shape[0]:
            raise ValueError(f"x must be 2D with shape (N, {kappa.shape[0]}), but got shape {x.shape}")
        return 0.5 * np.sum(kappa * np.square(x - xstar), axis=1)
    
    return potential

def linear_DD(phi):
    """
    Returns vectorized umbrella potential
    U(x) = sum(phi_i * x_i)
    phi has shape (D,)
    x has shape (N, D)
    U(x) has shape (N,)
    """
    phi = np.asarray(phi).reshape(-1)
    
    def potential(x):
        if x.ndim != 2 or x.shape[1] != phi.shape[0]:
            raise ValueError(f"x must be 2D with shape (N, {phi.shape[0]}), but got shape {x.shape}")
        return np.dot(x, phi)
    
    return potential

def morse(D, a, r_e):
    """
    Returns vectorized Morse potential
    U(r) = D * (1 - exp(-a*(r - r_e)))^2
    D: well depth
    a: controls width of potential
    r_e: equilibrium bond distance
    """
    return lambda r: D * np.square(1 - np.exp(-a * (r - r_e)))

def lennard_jones(epsilon, sigma):
    """
    Returns vectorized Lennard-Jones potential
    U(r) = 4 * epsilon * ((sigma/r)^12 - (sigma/r)^6)
    epsilon: well depth
    sigma: distance at which potential is zero
    """
    return lambda r: 4 * epsilon * (np.power(sigma/r, 12) - np.power(sigma/r, 6))

def gaussian(A, sigma):
    """
    Returns vectorized Gaussian potential
    U(x) = A * exp(-x^2 / (2*sigma^2))
    A: amplitude of the Gaussian
    sigma: standard deviation
    """
    return lambda x: A * np.exp(-np.square(x) / (2 * sigma**2))

def double_well(a, b):
    """
    Returns vectorized double well potential
    U(x) = x^4 - a*x^2 + b*x
    a, b: parameters controlling the shape of the double well
    """
    return lambda x: np.power(x, 4) - a * np.square(x) + b * x

def periodic(k, L):
    """
    Returns vectorized periodic potential
    U(x) = k * (1 - cos(2*pi*x/L))
    k: strength of the potential
    L: periodicity
    """
    return lambda x: k * (1 - np.cos(2 * np.pi * x / L))

# Función de utilidad para graficar potenciales
def plot_potential(potential_func, x_range, label):
    import matplotlib.pyplot as plt
    x = np.linspace(*x_range, 1000)
    y = potential_func(x)
    plt.plot(x, y, label=label)
    plt.xlabel('x')
    plt.ylabel('U(x)')
    plt.legend()
    plt.title('Potential Energy')
    plt.grid(True)


import importlib
import sys
from functools import lru_cache
from pathlib import Path

class AtomPositionLoader:
    """
    Lazily loads structural_file_readers modules on demand
    and dispatches read/export calls via the `.read()` and `.export()` methods.

    The `.read(source, file_location)` method auto-detects the source format
    by extension if `source` is None. After reading, the loader populates protected
    `_attr` fields and only publicly sets attributes that are not defined as properties.
    """

    def __init__(self, file_location: str = None, name: str = None, **kwargs):
        self.name = name
        self.file_location = file_location
        self._reader = None

        # Mapping: format key -> (class_name, export_method, read_method)
        self._source_map = {
            'CIF':    ('CIF',    'export_as_CIF',   'read_CIF'),
            'POSCAR': ('POSCAR', 'export_as_POSCAR', 'read_POSCAR'),
            'XYZ':    ('XYZ',    'export_as_xyz',    'read_XYZ'),
            'SI':     ('SI',     'export_as_SI',     'read_SI'),
            'PDB':    ('PDB',    'export_as_PDB',    'read_PDB'),
            'ASE':    ('ASE',    'export_as_ASE',    'read_ASE'),

            'AIMS':   ('AIMS',   'export_as_AIMS',   'read_AIMS'),
            'GEN':    ('GEN',    'export_as_GEN',    'read_GEN'),
            'DUMP':   ('DUMP',   'export_as_DUMP',   'read_DUMP'),
            'LAMMPS': ('LAMMPS','export_as_LAMMPS', 'read_LAMMPS'),
        }
        # File extension -> format key map
        self._ext_map = {
            'cif': 'CIF', 'poscar': 'POSCAR', 'vasp': 'POSCAR',
            'xyz': 'XYZ', 'si': 'SI', 'pdb': 'PDB', 'ase': 'ASE',
            'aims': 'AIMS', 'gen': 'GEN', 'dump': 'DUMP',
            'lmp': 'LAMMPS', 'lammpstrj': 'LAMMPS'
        }

    @lru_cache(maxsize=None)
    def _get_reader_class(self, source: str):
        key = source.upper()
        if key not in self._source_map:
            raise ValueError(f"Unknown source format '{source}'")
        class_name, _, _ = self._source_map[key]
        module_path = (
            "sage_lib.IO.structure_handling_tools.structural_file_readers."
            f"{class_name}"
        )
        try:
            module = importlib.import_module(module_path)
        except ImportError as e:
            sys.stderr.write(f"Error importing module {module_path}: {e}\n")
            return None
        if not hasattr(module, class_name):
            sys.stderr.write(f"Module '{module_path}' has no class '{class_name}'\n")
            return None
        return getattr(module, class_name)

    # ------------------------------------------------------------------
    def _inject_state(self, target):
        """
        Copy all loader fields to *target*.
        For names that start with “_”, first try the public
        property (self.attr) and pass that value on.
        Both aliases (_attr / attr) end up in *target*.
        """
        skip = {'_reader', 'name', 'file_location'}

        for key, val in self.__dict__.items():
            if key in skip or key.startswith('__'):
                continue

            if key.startswith('_') and key[1:] in [
                'uniqueAtomLabels', 
                'atomCount', 
                'atomLabelsList', 
                'atomPositions', 
                'latticeVectors',
                'timestep',
                'scaleFactor',
                'selectiveDynamics',
                'atomCoordinateType',
                'atomicConstraints',
                'latticeParameters',
                'latticeAngles',
                'E',
                'time',
                ]:

                pub = key[1:]
                # try self.attr (may trigger a property getter)
                val = getattr(self, pub, val)
                try:
                    setattr(target, pub, val)          # via setter if any
                except (AttributeError, TypeError):
                    target.__dict__[pub] = val         # read-only property
                target.__dict__[key] = val             # keep _attr alias
            else:
                # plain name first
                try:
                    setattr(target, key, val)
                except (AttributeError, TypeError):
                    target.__dict__[key] = val
                # add the protected twin
                target.__dict__.setdefault(f'_{key}', val)

        if isinstance(target, self._get_reader_class('PDB')):
            target.connection_list = self.get_connection_list(periodic=False)

    def read(self, source: str | None = None, file_location: str | None = None, **kwargs):
        """
        Detect format (if `source` is None), load file, cache the reader,
        and expose its attributes on the loader.
        """
        file_loc = file_location or self.file_location
        if not file_loc:
            raise ValueError("file_location must be specified for reading.")

        fmt_key = (source.upper() if source
                   else self._ext_map.get(Path(file_loc).suffix.lstrip('.').lower()))
        if not fmt_key:
            raise ValueError(f"Cannot detect format for '{file_loc}'")

        ReaderClass = self._get_reader_class(fmt_key)
        if ReaderClass is None:
            raise RuntimeError(f"Failed to import reader class for '{fmt_key}'")

        reader = ReaderClass(name=self.name, file_location=file_loc)
        _, _, read_method = self._source_map[fmt_key]
        if not hasattr(reader, read_method):
            raise AttributeError(f"{ReaderClass.__name__} lacks '{read_method}'")

        result = getattr(reader, read_method)(file_loc, **kwargs)

        # cache & expose reader’s state
        self._reader = reader
        for k, v in reader.__dict__.items():
            setattr(self, f"_{k}", v)
            if not isinstance(getattr(type(self), k, None), property):
                setattr(self, k, v)

        return result

    def export(self, fmt: str, file_location: str | None = None, **kwargs):
        """
        Create a fresh writer of type `fmt`, inject current state, call its exporter.
        """
        file_loc = file_location or self.file_location
        fmt_key = fmt.upper()

        WriterClass = self._get_reader_class(fmt_key)
        if WriterClass is None:
            raise ValueError(f"Writer class for '{fmt_key}' not found.")

        writer = WriterClass(name=self.name, file_location=file_loc)
        self._inject_state(writer)
        export_method = self._source_map[fmt_key][1]
        if not hasattr(writer, export_method):
            raise AttributeError(f"{WriterClass.__name__} lacks '{export_method}'")

        return getattr(writer, export_method)(file_loc, **kwargs)

    # ------------------------------------------------------------------
    # dynamic helpers (read_<fmt>, export_as_<fmt>, export_<fmt>)
    # ------------------------------------------------------------------
    def __getattr__(self, attr):
        # ------ read_<FMT> -------------------------------------------------
        if attr.startswith('read_'):
            fmt_key = attr.split('_', 1)[1].upper()
            if fmt_key in self._source_map:
                def _reader(file_location=None, **kw):
                    res = self.read(fmt_key, file_location=file_location, **kw)
                    #if hasattr(self, 'group_elements_and_positions'):
                    #    self.group_elements_and_positions()
                    return res
                return _reader

        # ------ export_as_<FMT> -------------------------------------------
        if attr.startswith('export_as_'):
            fmt_key = attr.split('_', 2)[2].upper()
            if fmt_key in self._source_map:
                def _exporter(file_location=None, **kw):
                    return self.export(fmt_key, file_location=file_location, **kw)
                return _exporter

        # ------ export_<FMT> ----------------------------------------------
        if attr.startswith('export_') and not attr.startswith('export_as_'):
            fmt_key = attr.split('_', 1)[1].upper()
            if fmt_key in self._source_map:
                def _exp(file_location=None, **kw):
                    return self.export(fmt_key, file_location=file_location, **kw)
                return _exp

        raise AttributeError(f"{self.__class__.__name__} has no attribute '{attr}'")




try:
    from ..master.FileManager import FileManager
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing FileManager: {str(e)}\n")
    del sys

try:
    import numpy as np
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing numpy: {str(e)}\n")
    del sys

try:
    import subprocess
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing subprocess: {str(e)}\n")
    del sys
    
class BashScriptManager(FileManager):
    def __init__(self, file_location:str=None, name:str=None, **kwargs):
        super().__init__(file_location=file_location, name=name)
        self._comment = None
        self._parameters = {'srun':[], 'module':[], 'export':[] }
        self._BatchFile = None

        self._parameter_descriptions = {
        "--nodes": "Number of nodes to be used in the job.",
        "--ntasks-per-node": "Number of tasks to run per node.",
        "--time": "Wall clock time limit for the job.",
        "-o": "File to which standard output will be redirected.",
        "-e": "File to which standard error will be redirected.",
        "--mem": "Memory to be allocated for the job.",
        "--array": "Specifies the indexes of the job array elements.",
        "--mail-type": "Event types to notify the user.",
        "--mail-user": "Email address for job notifications.",
        "--cpus-per-task": "Number of CPU cores to be allocated per task.",
        "--ntasks": "Total number of tasks across all nodes.",
        "--ntasks-per-core": "Number of tasks to be launched per core.",
        "--constraint": "Node feature constraints (e.g., gpu).",
        "--gres": "Generic resources (e.g., GPUs) to allocate.",
                                        }
        
    def readBashScript(self, file_location:str=None):
        file_location = file_location if type(file_location) == str else self.file_location
        lines =list(self.read_file(file_location,strip=False))
        self.BatchFile = lines

        for line in lines:
            line = line.strip()
            if line.startswith("#SBATCH"):
                param = line.split(" ", 1)[1]
                self._parameters[param.split("=")[0]] = param.split("=")[1] if "=" in param else None
            if line.startswith("export"):
                self._parameters['export'].append( line[7:] )
            if line.startswith("srun"):
                self._parameters['srun'].append( line[5:] )
            if line.startswith("module"):
                self._parameters['module'].append( line[7:] )
            elif line.startswith("#"):
                self._comment = line[1:].strip()

    def exportAsBash(self, file_location:str=None):
        # Determine the file location
        file_location = file_location if file_location is not None else self.file_location

        # Open the file in write mode
        with open(file_location, 'w') as f:
            for line in self.BatchFile:
                f.write(line)
        # Give execute permission to the VASPscript.sh file
        subprocess.run(['chmod', '+x', file_location])

    def makeBash(self, file_location=None, job_name='SAGE_Bash_Script', output_file='SAGE_OutPut.out', error_file='SAGE_error.err', 
                        nodes:int=None, tasks_per_node:int=None, time_limit=None, conda_env=None, commands=None):
        file_location = file_location if file_location is not None else self.file_location if self.file_location is not None else '.'
        conda_env = conda_env if conda_env is not None else []
        commands = commands if commands is not None else []
        
        with open(script_filename, 'w') as file:
            file.write("#!/bin/bash\n")
            file.write(f"#SBATCH --job-name={job_name}\n")
            file.write(f"#SBATCH --output={output_file}\n")
            file.write(f"#SBATCH --error={error_file}\n")
            file.write(f"#SBATCH --nodes={nodes}\n")
            file.write(f"#SBATCH --ntasks-per-node={tasks_per_node}\n")
            file.write(f"#SBATCH --time={time_limit}\n")
            file.write("\n")
            
            file.write("# Activate the Python environment\n")
            for env in conda_env:
                file.write(f"source activate {env}\n")
            
            file.write("\n")
            for command in commands:
                file.write(f"{command}\n")

    def summary(self):
        text_str = "Batch Script Summary:\n"
        for param, value in self._parameters.items():
            if not type(value) == list:
                description = self.parameter_descriptions.get(param, "No description available.")
                text_str += f"{str(param)[:15].ljust(10, ' ') }: {str(value)[:10].ljust(10, ' ') } ({description})\n"
        return text_str

'''
pot = BashScriptManager('/home/akaris/Documents/code/Physics/VASP/v6.1/files/BASH/QUIP.sh')
pot.readBashScript()

pot.exportAsBash('/home/akaris/Documents/code/Physics/VASP/v6.1/files/bulk_optimization/Pt/FCC100/script1.sh')
print(pot.parameters)
print( pot.summary() )
pot.exportAsPOTCAR('/home/akaris/Documents/code/Physics/VASP/v6.1/files/POTCAR/POTCAR_export')
'''




try:
    from .PeriodicSystem import PeriodicSystem
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing sage_lib.IO.structure_handling_tools.PeriodicSystem: {str(e)}\n")
    del sys

try:
    from .NonPeriodicSystem import NonPeriodicSystem
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing sage_lib.IO.structure_handling_tools.NonPeriodicSystem: {str(e)}\n")
    del sys

try:
    from .plot import plot
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing sage_lib.IO.structure_handling_tools.NonPeriodicSystem: {str(e)}\n")
    del sys

try:
    import numpy as np
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing numpy: {str(e)}\n")
    del sys

class AtomPosition(PeriodicSystem, NonPeriodicSystem, plot):
    def __init__(self, file_location:str=None, name:str=None, **kwargs):
        PeriodicSystem.__init__(self, name=name, file_location=file_location)
        NonPeriodicSystem.__init__(self, name=name, file_location=file_location)
        self._comment = None
        self._atomCount = None  # N total number of atoms
        self._data_dtype = np.float64
        
    def NonPeriodic_2_Periodic(self, latticeVectors:np.array=None, center:bool=True, vacuum:float=12.0):
        """
        Converts a NonPeriodicSystem instance to a PeriodicSystem instance by creating a new PeriodicSystem 
        object and copying shared attributes from the NonPeriodicSystem. It then adjusts the atom positions 
        to ensure that the atom with the lowest coordinate in each axis is at least 6 Angstroms from the origin.
        Finally, it sets the lattice vectors based on the adjusted atom positions.

        :return: A new instance of PeriodicSystem with shared attributes copied and adjusted atom positions.
        """ 
        if self.latticeVectors is not None: 
            return self

        # Adjust atom positions
        min_coords = np.min(self.atomPositions, axis=0)
        max_coords = np.max(self.atomPositions, axis=0)
 
        self.atomPositions += (vacuum/2 - min_coords) if center else 0

        # Set lattice vectors
        self.latticeVectors = latticeVectors if latticeVectors is not None else \
            np.array([  [max_coords[0] + vacuum/2, 0, 0], 
                        [0, max_coords[1] + vacuum/2, 0], 
                        [0, 0, max_coords[2] + vacuum/2]])

        self.atomCoordinateType = 'Cartesian'

        return self



try:
    from ..master.FileManager import FileManager
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing FileManager: {str(e)}\n")
    del sys

# Defining the ChargeFileManager class which extends from FileManager.
class ChargeFileManager(FileManager):
    """
    ChargeFileManager Class

    A specialized file manager designed to handle computational charge files,
    particularly useful for VASP-related operations. It includes advanced
    functionalities like efficient copying of large files and storing the path
    of original files for reference.

    Attributes:
        file_location (str): The location of the file being managed.
        name (str): The name of the file or the identifier.
        data: An attribute to store additional data if required.
    """

    def __init__(self, file_location: str = None, name: str = None, **kwargs):
        """
        Initialize the ChargeFileManager instance.

        Args:
            file_location (str, optional): Initial location of the file to manage.
            name (str, optional): Name or identifier for the file.
            **kwargs: Arbitrary keyword arguments for future extensions.
        """
        # Initialize the parent class FileManager with provided arguments.
        super().__init__(name=name, file_location=file_location)
        # Additional data attribute for storing any related data.
        self.data = None

    def exportFile(self, file_location: str = None, original_file_location: str = None) -> bool:
        """
        Export the managed file to a specified location.

        This method copies the file from its original location to a new location,
        providing a way to effectively distribute or backup the file.

        Args:
            file_location (str, optional): The destination location to copy the file.
            original_file_location (str, optional): The original location of the file,
                defaults to the current file location stored in the instance.

        Returns:
            bool: True if the operation is successful, otherwise False.
        """
        # Determine the original file location.
        original_file_location = original_file_location if original_file_location is not None else self.file_location

        # Perform the file copy operation.
        self.copy(file_location, original_file_location)

        return True

    def importFileLocation(self, file_location:str = None) -> bool:
        """
        Update the file location in the instance.

        This method allows updating the file location, enabling the management
        of different files or changing the reference to the current file.

        Args:
            file_location (str, optional): The new location of the file to manage.

        Returns:
            bool: True if the operation is successful, otherwise False.
        """
        # Update the file_location attribute with the new file location.
        file_location = file_location if type(file_location) == str else self.file_location
        self._file_location = file_location

        return True



try:
    from ...master.FileManager import FileManager
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing FileManager: {str(e)}\n")
    del sys

try:
    import numpy as np
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing numpy: {str(e)}\n")
    del sys

class InputFileManager(FileManager):
    def __init__(self, file_location:str=None, name:str=None, **kwargs):
        super().__init__(name=name, file_location=file_location)
        self._comment = None

    def var_assing(self, var_name, var_value):
        if not var_name in self.attr_dic.keys():
            self.attr_dic[var_name] = var_value
            setattr(self, var_name, var_value)
        else:
            self.attr_dic[var_name] = var_value
            setattr(self, var_name, var_value)
            



from setuptools import setup, find_packages

setup(
    name='sage_lib',  # El nombre del paquete
    version='0.1.6.47',
    description='A library for advanced scientific calculations and visualization',
    long_description=open('README.md', encoding='utf-8').read(),  # Updated to specify encoding for better compatibility
    long_description_content_type='text/markdown',
    author='Lombardi, Juan Manuel',
    author_email='lombardi@fhi-berlin.mpg.de',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    license='MIT',
    install_requires=[
        'numpy>=1.21.0',  # Updated to a more recent version for better compatibility
        'matplotlib>=3.4.0',
        'scipy>=1.7.0',
        'imageio>=2.10.0',
        'tqdm>=4.50.0',
        'joblib>=1.0.0',
        'seaborn>=0.11.0',
        'scikit-learn>=0.24.0',
        'umap-learn>=0.5.0',
        'kneed>=0.7.0',
        'hdbscan>=0.8.0',
        'dscribe',
    	'h5py>=3.6.0',
	],
    extras_require={
        'dev': [
            'pytest>=6.0',  # Specified minimum version for compatibility
            'flake8>=3.8',
            'black',  # Added black for code formatting
            'mypy',  # Added mypy for type checking
            'coverage',  # Added for code coverage analysis
            'tox',  # Added for testing across multiple environments
        ],
        'docs': [
            'sphinx>=4.0',
            'sphinx-rtd-theme>=0.5.0',
            'myst-parser',  # Added MyST parser for enhanced Markdown support in Sphinx
            'sphinx-autodoc-typehints',  # Added for better type hint documentation
        ],
        'ml': [
            'scikit-learn>=0.24.0',
            'umap-learn>=0.5.0',
            'hdbscan>=0.8.0',
        ],
        'viz': [
            'matplotlib>=3.4.0',
            'seaborn>=0.11.0',
        ],
    },
    entry_points={
        'console_scripts': [
            'sage = sage_lib.main:main',  # Establece 'sage' como el comando de consola
        ],
    },
    keywords='scientific calculations visualization computational chemistry quantum mechanics machine learning data analysis',  # Expanded keywords for better discoverability
    #project_urls={
    #    'Bug Reports': '',
    #    'Source': '',
    #    'Documentation': '',  # Added a link to documentation
    #},
)



import itertools
from typing import Tuple, Union

import numpy as np
import torch
from torch import nn, Tensor
from torch.autograd.functional import jacobian

__all__ = [
    "WeightFunc2BodyIdentity",
    "WeightFunc2BodyInvDist",
    "WeightFunc3BodyUnity",
    "WeightFunc3BodyExpInvDist",
    "GeomFunc2BodyInvDist",
    "GeomFunc3BodyAngle",
    "GeomFunc3BodyCosAngle",
    "DistFuncGaussian",
    "mbtr_python",
]


class ScalarFunc(nn.Module):
    """
    Base class for all scalar functions (geometry and weight functions).
    If an algebraic derivative function is not defined, we use PyTorch's
    auto-differentiation function to calculate a derivative value.
    """

    k = 1

    def _check_r_shape(self, r):
        if len(r.shape) != 3:
            raise ValueError("r must have a shape of [batch_size, k=%d, xyz]" % self.k)
        if r.shape[1] != self.k:
            raise ValueError(
                "The second dim of r must be %d, got %d" % (self.k, r.shape[1])
            )
        if r.shape[2] != 3:
            raise ValueError(
                "Expected %d Cartesian coordinates, got %d" % (3, r.shape[2])
            )

    def forward(self, r):
        self._check_r_shape(r)
        return self._forward(r)

    def div(self, r):
        """Fallback to PyTorch auto-differentiation if not implemented."""
        self._check_r_shape(r)

        try:
            return self._div(r)
        except NotImplemented:
            return jacobian(lambda x: self(x).sum(dim=0), r, vectorize=True)

    def _forward(self, r):
        raise NotImplemented()

    def _div(self, r):
        raise NotImplemented()


class WeightFunc2BodyIdentity(ScalarFunc):
    """
    2-body weighting function of constant 1.
    """

    k = 2

    def _forward(self, r):
        return torch.ones_like(r[:, 0, 0])

    def _div(self, r):
        return torch.zeros_like(r)


class WeightFunc2BodyInvDist(ScalarFunc):
    """
    2-body weighting function that takes the form of:

    .. math::
        w(R_i, R_j) = \\frac{1}{|R_i - R_j|^2}
    """

    k = 2

    def _forward(self, r):
        return 1.0 / torch.linalg.norm(r[:, 0] - r[:, 1], dim=-1) ** 2

    def _div(self, r):
        diff = r[:, 0] - r[:, 1]
        norm = torch.linalg.norm(diff, dim=-1, keepdim=True)
        d = -2.0 / norm**4 * diff
        return torch.cat([d.unsqueeze(1), -d.unsqueeze(1)], dim=1)


class WeightFunc3BodyUnity(ScalarFunc):
    """
    3-body weighting function that always returns 1.0:
    """

    k = 3

    def _forward(self, r):
        return torch.ones_like(r[:, 0, 0])

    def _div(self, r):
        return torch.zeros_like(r)


class WeightFunc3BodyExpInvDist(ScalarFunc):
    """
    3-body weighting function that takes the form of:

    .. math::
        w(R_i, R_j, R_k) = exp\\left(-\\frac{|R_i-R_j|+|R_j-R_k|+|R_k-R_i|}{ls}
        \\right)
    """

    k = 3

    def __init__(self, ls):
        super().__init__()
        self.ls = ls

    def _forward(self, r):
        ri, rj, rk = r[:, 0], r[:, 1], r[:, 2]
        norms = (
            torch.linalg.norm(ri - rj, dim=-1)
            + torch.linalg.norm(ri - rk, dim=-1)
            + torch.linalg.norm(rj - rk, dim=-1)
        )
        return torch.exp(-norms / self.ls)


class GeomFunc2BodyInvDist(ScalarFunc):
    """
    Geometry function that takes the form of:

    .. math::
        G(R_i, R_j) = \\frac{1}{|R_i - R_j|}
    """

    k = 2

    def _forward(self, r):
        print(r)
        return 1.0 / torch.linalg.norm(r[:, 0] - r[:, 1], dim=-1)

    def _div(self, r):
        diff = r[:, 0] - r[:, 1]
        #print(r) ### distancia
        norm = torch.linalg.norm(diff, dim=-1, keepdim=True)
        d = -1.0 / norm**3 * diff
        return torch.cat([d.unsqueeze(1), -d.unsqueeze(1)], dim=1)


class GeomFunc3BodyAngle(ScalarFunc):
    """
    Geometry function of angle function:

    .. math::
        G(R_i, R_j, R_k) = \\arccos(\\frac{(R_i-R_j)\\cdot(R_k-R_j)}{|R_i-R_j|
        \\cdot|R_k-R_j|})
    """

    k = 3

    def _forward(self, r):
        ra, rb, rc = r[:, 0], r[:, 1], r[:, 2]
        dotuv = torch.sum((ra - rb) * (rc - rb), dim=-1)
        denominator = torch.linalg.norm(ra - rb, dim=-1) * torch.linalg.norm(
            rc - rb, dim=-1
        )
        return torch.arccos(torch.clamp(dotuv / denominator, min=-1.0, max=1.0))

    def _div(self, r):
        ra, rb, rc = r[:, 0], r[:, 1], r[:, 2]
        vab = ra - rb
        vcb = rc - rb

        dotuv = torch.sum(vab * vcb, dim=-1)
        dab = torch.linalg.norm(vab, dim=-1)
        dcb = torch.linalg.norm(vcb, dim=-1)
        cos = dotuv / (dab * dcb)

        factor = -1 / torch.sqrt(1 - torch.clamp(cos**2, min=0.0, max=1 - 1e-7))
        da = (dab**2 * vcb - dotuv * vab) / (dab**3 * dcb)
        db = (
            dab * dcb * (-vab - vcb) + dotuv * (dcb / dab * vab + dab / dcb * vcb)
        ) / (dab**2 * dcb**2)
        dc = (dcb**2 * vab - dotuv * vcb) / (dcb**3 * dab)

        return factor * torch.cat(
            [da.unsqueeze(1), db.unsqueeze(1), dc.unsqueeze(1)], dim=1
        )


class GeomFunc3BodyCosAngle(ScalarFunc):
    """
    Geometry function of cosine function:

    .. math::
        G(R_i, R_j, R_k) = \\frac{(R_i-R_j)\\cdot(R_k-R_j)}{|R_i-R_j|
        \\cdot|R_k-R_j|}
    """

    k = 3

    def _forward(self, r):
        ra, rb, rc = r[:, 0], r[:, 1], r[:, 2]
        dotuv = torch.tensordot(ra - rb, rc - rb, dims=([1], [1]))
        denominator = torch.linalg.norm(ra - rb, dim=-1) * torch.linalg.norm(
            rc - rb, dim=-1
        )
        return torch.maximum(-1.0, torch.minimum(1.0, dotuv / denominator))


class DistFuncGaussian(nn.Module):
    """
    Gaussian distribution function.
    """

    def __init__(self, sigma):
        super().__init__()
        self.const = float(1.0 / (sigma * np.sqrt(2.0)))

    def _check_shapes(self, val_range, geom_mean):
        assert len(val_range.shape) == 1, "val_range must be a 1d tensor."
        assert (
            len(geom_mean.shape) == 1
        ), "geom_mean must be a 1d tensor with first dim as batch size."

    def forward(self, val_range, geom_mean, *, dx):
        self._check_shapes(val_range, geom_mean)

        val_range = val_range.unsqueeze(0)
        geom_mean = geom_mean.unsqueeze(1)
        right = torch.erf((val_range + dx - geom_mean) * self.const)
        left = torch.erf((val_range - geom_mean) * self.const)
        return (right - left) / 2

    def div(self, val_range: Tensor, geom_mean: Tensor, *, dx):
        self._check_shapes(val_range, geom_mean)

        val_range = val_range.unsqueeze(0)
        geom_mean = geom_mean.unsqueeze(1)

        zr: Tensor = (val_range + dx - geom_mean) * self.const
        zl: Tensor = (val_range - geom_mean) * self.const

        return (-self.const / np.sqrt(np.pi)) * (
            torch.exp(-(zr**2)) - torch.exp(-(zl**2))
        )


class MBTR(nn.Module):
    def __init__(self, order, geomf, weightf, distf, grid):
        super().__init__()

        self.geomf = geomf
        self.weightf = weightf
        self.distf = distf

        self.order = order
        self.grid = nn.Parameter(torch.tensor(grid), requires_grad=False)

    def forward(self, r, z, compute_div=True):
        # z: torch.tensor((n_atoms,))
        # r: torch.tensor((batch, n_atoms, 3))
        elements = sorted(set(z))
        b_size = r.shape[0]
        grid_size = self.grid.size(0)
        coord_size = r.shape[2]
        n_atom = r.shape[1]
        dx = self.grid[1] - self.grid[0]
        full = slice(None, None, None)

        mbtr_shape = (b_size,) + (len(elements),) * self.order + (grid_size,)

        mbtr = torch.zeros(mbtr_shape, dtype=r.dtype, device=self.grid.device)
        mbtr_div = (
            torch.zeros(
                mbtr_shape + (n_atom, coord_size),
                dtype=r.dtype,
                device=self.grid.device,
            )
            if compute_div
            else None
        )

        for atom_ids in itertools.product(*([range(n_atom)] * self.order)):
            if len(set(atom_ids)) != len(atom_ids):
                # Contains duplicates
                continue

            zs = [elements.index(z[x]) for x in atom_ids]
            rs = r[:, atom_ids]

            gf = self.geomf(rs)
            wf = self.weightf(rs).unsqueeze(1)
            grid_values = self.distf(self.grid, gf, dx=dx)

            indexer = (full,) + tuple(zs) + (full,)
            mbtr[indexer] += wf * grid_values

            if compute_div:
                wf_div = self.weightf.div(rs)
                gf_div = self.geomf.div(rs)
                grid_div = self.distf.div(self.grid, gf, dx=dx)

                div = (
                    torch.bmm(
                        grid_values.unsqueeze(2),
                        wf_div.reshape((b_size, -1)).unsqueeze(1),
                    )
                    + torch.bmm(
                        grid_div.unsqueeze(2), gf_div.reshape((b_size, -1)).unsqueeze(1)
                    )
                    * wf.unsqueeze(2)
                ).reshape((b_size, grid_size, self.order, coord_size))
                mbtr_div[(indexer + (atom_ids, full))] += div

        if compute_div:
            return mbtr, mbtr_div
        return mbtr


@torch.no_grad()
def mbtr_python(
    z: "np.ndarray",
    r: "np.ndarray",
    grid: Union["np.ndarray", Tuple],
    order: int,
    weightf: nn.Module,
    distf: nn.Module,
    geomf: nn.Module,
    flatten=False,
    device="cpu",
    as_numpy=True,
) -> Tuple["np.ndarray", "np.ndarray"]:
    """
    Compute MBTR using PyTorch tensor library.

    :param z: Element types. (NAtom)
    :param r: Coordinates of atoms. (Batch, NAtom, 3)
    :param order: Order of MBTR.
    :param weightf: Weighting function.
    :param distf: Distribution function.
    :param geomf: Geometry function.
    :param grid: Grid definition.
    :param flatten: Whether to flatten representation.
    :param device: Device.
    :param as_numpy: As numpy array.
    :return: MBTR tensor.
    """
    if isinstance(grid, tuple):
        grid = np.linspace(*grid)

    model = MBTR(
        order=order,
        weightf=weightf,
        distf=distf,
        geomf=geomf,
        grid=grid,
    ).to(device)

    torch_r = torch.tensor(r).to(device)
    rep, rep_div = model(torch_r, z, compute_div=True)

    if as_numpy:
        rep, rep_div = rep.cpu().numpy(), rep_div.cpu().numpy()

    if flatten:
        return (
            rep.reshape((len(rep), -1)),
            rep_div.reshape((len(rep_div), -1, r.shape[1], r.shape[2])),
        )
    return rep, rep_div



'''
from mbtr_grad.mbtr_python_torch import (
    mbtr_python,
    WeightFunc2BodyInvDist,
    GeomFunc2BodyInvDist,
    DistFuncGaussian,
)
'''

aspirin = {
    'z': np.asarray([6, 6, 6, 6, 6, 6, 6, 8, 8, 8, 6, 6, 8, 1, 1, 1, 1, 1, 1, 1, 1]),
    'r': np.asarray([
        [+1.967773, -1.138310, -0.543556],
        [+0.645441, +0.992465, -1.702570],
        [+2.234543, -0.777031, -1.830163],
        [+1.581020, +0.289498, -2.446978],
        [-3.016591, +1.741695, +0.612453],
        [+0.878228, -0.578833, +0.189787],
        [+0.263598, +0.519752, -0.384926],
        [+1.503426, -1.267264, +2.463581],
        [-2.236631, +0.279797, -1.121948],
        [-0.715723, -1.175512, +1.929332],
        [+0.633549, -0.964428, +1.667591],
        [-2.016979, +0.998229, -0.174489],
        [-0.753226, +1.137627, +0.363041],
        [-0.693787, -1.606394, +2.839317],
        [+2.522198, -1.999940, -0.132890],
        [+0.118692, +1.850180, -2.156963],
        [+3.099397, -1.279973, -2.334072],
        [+1.852263, +0.669541, -3.409804],
        [-4.006237, +1.796931, +0.063728],
        [-3.090694, +1.158956, +1.570798],
        [-2.624328, +2.785921, +0.842982] 
    ])
}

aspirin_range = np.linspace(0, 1.1, 500)
aspirin_rep, aspirin_rep_div = mbtr_python(
    z=aspirin['z'], r=aspirin['r'][None],
    order=2, grid=aspirin_range,
    weightf=WeightFunc2BodyInvDist(), 
    geomf=GeomFunc2BodyInvDist(),
    distf=DistFuncGaussian(2**-4.5)
)

print(aspirin_rep.shape, aspirin_rep_div.shape)
import matplotlib.pyplot as plt 
plt.plot( aspirin_rep[0,0,:,:].T )
plt.show()
plt.plot( aspirin_rep[0,1,:,:].T )
plt.show()
plt.plot( aspirin_rep[0,2,:,:].T )
plt.show()
# Should display:
# (1, 3, 3, 500) (1, 3, 3, 500, 21, 3)


import numpy as np
import itertools
from scipy.spatial import cKDTree

class PeriodicCKDTree(cKDTree):
    """
    cKDTree subclass supporting periodic boundary conditions.

    Behaviors:
      - Orthorhombic box (1D bounds) with full periodicity: uses native cKDTree methods.
      - General box (2D bounds matrix): uses query tiling to implement PBC.
    """
    def __init__(self, bounds, data, leafsize=10, pbc=None, force_orth:bool=False):
        data = np.asarray(data, float)
        d    = data.shape[1]

        # Normalize pbc
        if pbc is None:
            pbc = (True,) * d
        elif len(pbc) != d:
            raise ValueError(f"pbc must have length {d}")
        self.pbc = tuple(pbc)

        # Force float-array form
        # Ensure bounds is a float array
        bounds = np.asarray(bounds, float)

        # Determine if we can use the native orthorhombic periodic support
        is_orth = force_orth and ((bounds.ndim==1 and bounds.size==d) \
                  or (bounds.ndim==2 and bounds.shape==(d,d) \
                      and not np.any(np.abs(bounds[~np.eye(d, dtype=bool)])>1e-12)))

        if is_orth and all(self.pbc):
            # Use native cKDTree periodic support for orthorhombic box
            # Orthorhombic periodic box: use native cKDTree periodic support
            box = bounds if bounds.ndim == 1 else np.diag(bounds)
            super().__init__(data, leafsize=leafsize, boxsize=box)
            self._use_native = True
            self.bounds = np.diag(box)
        else:
            # Fallback: plain cKDTree + manual tiling
            # General case: build plain cKDTree on data (no tiling)
            super().__init__(data, leafsize=leafsize)
            self._use_native = False
            if bounds.ndim == 1:
                self.bounds = np.diag(bounds)
            elif bounds.ndim == 2 and bounds.shape == (d, d):
                self.bounds = bounds
            else:
                raise ValueError(f"bounds must be length-{d} or {d}x{d} matrix")

        self._n_orig = data.shape[0]

    def __reduce__(self):
        fn, args, state = super().__reduce__()  
        extra = (self._use_native, self.pbc, self.bounds, self._n_orig)
        return (self.__class__._rebuild, (fn, args, state) + extra)

    @staticmethod
    def _rebuild(fn, args, state, use_native, pbc, bounds, n_orig):
        tree = fn(*args)
        super(PeriodicCKDTree, tree).__setstate__(state)
        tree._use_native = use_native
        tree.pbc          = pbc
        tree.bounds       = bounds
        tree._n_orig      = n_orig
        return tree
    
    @property
    def use_native(self) -> bool:
        """
        Whether this tree is using SciPy’s native periodic-box support
        (True) or the custom tiling implementation (False).
        """
        return self._use_native

    @use_native.setter
    def use_native(self, flag: bool):
        """
        Manually enable or disable native-box support.

        Parameters
        ----------
        flag : bool
            True to force native periodic support; False to force the
            fallback tiling implementation.
        """
        self._use_native = bool(flag)

    def _make_shifts(self, r):
        # Compute integer shifts needed to cover radius r
        lengths = np.linalg.norm(self.bounds, axis=0)
        max_shifts = np.ceil(r / lengths).astype(int)
        axes = [range(-m, m + 1) if p else (0,)
                for m, p in zip(max_shifts, self.pbc)]
        return np.array(list(itertools.product(*axes)), int)

    def query(self, x, k=1, eps=0, p=2, distance_upper_bound=np.inf):
        if self.use_native:
            return super().query(x, k=k, eps=eps, p=p, distance_upper_bound=distance_upper_bound)

        x_arr = np.asarray(x, float)
        single = (x_arr.ndim == 1)
        Q = x_arr.reshape(-1, x_arr.shape[-1])
        d = Q.shape[1]

        # generate shifts for triclinic PBC: {-1,0,1} on periodic axes
        axes = [(-1, 0, 1) if p else (0,) for p in self.pbc]
        shifts_i = np.array(list(itertools.product(*axes)), dtype=int)
        shifts_r = shifts_i.dot(self.bounds)  # shape (S, d), lattice-vector shifts

        # tile queries and run a single kNN on the base tree
        tiledQ = (Q[:, None, :] + shifts_r[None, :, :]).reshape(-1, d)
        dists, idxs = super().query(
            tiledQ, k=k, eps=eps, p=p, distance_upper_bound=distance_upper_bound
        )

        # reshape to (nQ, S*k) and keep best k per original query
        S = shifts_r.shape[0]
        dists = dists.reshape(Q.shape[0], S * k)
        idxs  = idxs.reshape(Q.shape[0], S * k)

        # select top-k along axis=1
        part = np.argpartition(dists, kth=np.minimum(k-1, dists.shape[1]-1), axis=1)[:, :k]
        row_idx = np.arange(Q.shape[0])[:, None]
        d_best = dists[row_idx, part]
        i_best = idxs[row_idx, part]

        # sort those top-k to maintain ascending order
        order = np.argsort(d_best, axis=1)
        d_best = np.take_along_axis(d_best, order, axis=1)
        i_best = np.take_along_axis(i_best, order, axis=1)

        i_best = np.mod(i_best, self._n_orig)

        if single and k == 1:
            return d_best[0, 0], i_best[0, 0]
        if single:
            return d_best[0], i_best[0]
        if k == 1:
            return d_best[:, 0], i_best[:, 0]
        return d_best, i_best
        
    def query_ball_point(self, x, r, p=2., eps=0):
        if self.use_native:
            return super().query_ball_point(x, r, p, eps)

        x_arr = np.asarray(x, float)
        single = (x_arr.ndim == 1)
        Q = x_arr.reshape(-1, x_arr.shape[-1])

        shifts_i = self._make_shifts(r)
        shifts_r = shifts_i.dot(self.bounds)
        tiled = (Q[:, None, :] + shifts_r[None, :, :]).reshape(-1, Q.shape[1])
        raw = super().query_ball_point(tiled, r, p, eps)

        raw = np.array(raw, object).reshape(Q.shape[0], -1)
        out = []
        for row in raw:
            idxs = np.concatenate(row) % self._n_orig
            out.append( np.unique(idxs).astype(np.int64) )

        return out[0] if single else out

    def query_ball_tree(self, other, r, p=2., eps=0):
        if self.use_native and getattr(other, 'use_native', False):
            return super().query_ball_tree(other, r, p, eps)
        if not isinstance(other, PeriodicCKDTree):
            raise ValueError("Other tree must be PeriodicCKDTree")
        return other.query_ball_point(self.data, r, p, eps)

    def query_pairs(self, r, p=2., eps=0):
        if self.use_native:
            return super().query_pairs(r, p, eps)
        pairs = set()
        neighbors = self.query_ball_point(self.data, r, p, eps)
        for i, nbrs in enumerate(neighbors):
            for j in nbrs:
                if i < j:
                    pairs.add((i, j))
        return sorted(pairs)

    def count_neighbors(self, other, r, p=2.):
        '''
        if self._use_native and getattr(other, '_use_native', False):
            return super().count_neighbors(other, r, p)
        if not isinstance(other, PeriodicCKDTree):
            raise ValueError("Other tree must be PeriodicCKDTree")
        raw = super().count_neighbors(other, r, p)
        return np.array(raw).reshape(-1, self._n_orig).sum(axis=0)
        '''
        if self.use_native and getattr(other, 'use_native', False):
            return super().count_neighbors(other, r, p)
        if not isinstance(other, PeriodicCKDTree):
            raise ValueError("Other tree must be PeriodicCKDTree")

        counts = np.zeros(other.n, dtype=int)
        for i, point in enumerate(other.data):
            indices = self.query_ball_point(point, r, p)
            counts[i] = len(indices)
            
        return counts

    def sparse_distance_matrix(self, other, max_distance, p=2.):
        if self.use_native and getattr(other, 'use_native', False):
            return super().sparse_distance_matrix(other, max_distance, p)
        if not isinstance(other, PeriodicCKDTree):
            raise ValueError("Other tree must be PeriodicCKDTree")
        raw = super().sparse_distance_matrix(other, max_distance, p)
        result = {}
        for (i_t, j_t), dist in raw.items():
            i, j = i_t % self._n_orig, j_t % other._n_orig
            key = (i, j) if i < j else (j, i)
            if key not in result or dist < result[key]:
                result[key] = dist

        return result

    def sparse_distance_matrix2(self, other, max_distance, p=2.):
        """
        Periodic sparse distance matrix under a general lattice.

        Returns a dict mapping (i, j) -> distance for all pairs with
        minimum-image distance <= max_distance.

        Behavior:
          - If both trees use native orthorhombic periodic support, defer to SciPy.
          - Otherwise, tile 'other' by the lattice shifts sufficient to cover
            max_distance and compute a single sparse matrix query.
          - If self is other, exclude (i,i) and return only i<j, matching SciPy.

        Parameters
        ----------
        other : PeriodicCKDTree
            The other tree (must share the same dimensionality and lattice).
        max_distance : float
            Distance cutoff (must be finite).
        p : float
            Minkowski norm parameter (as in cKDTree).
        """
        # Fast path: both trees can rely on SciPy’s native orthorhombic support
        if self.use_native and getattr(other, 'use_native', False):
            return super().sparse_distance_matrix(other, max_distance, p)

        if not isinstance(other, PeriodicCKDTree):
            raise ValueError("Other tree must be PeriodicCKDTree")

        if not np.isfinite(max_distance):
            raise ValueError("max_distance must be finite for periodic tiling")

        # Basic sanity checks
        if self.data.shape[1] != other.data.shape[1]:
            raise ValueError("Dimensionality mismatch between trees")
        if self.bounds.shape != other.bounds.shape or not np.allclose(self.bounds, other.bounds, atol=1e-12):
            raise ValueError("Both trees must share the same lattice (bounds)")
        if self.pbc != other.pbc:
            raise ValueError("Both trees must share the same PBC tuple")

        # Compute lattice shifts to cover the search radius
        shifts_i = self._make_shifts(max_distance)              # (S, d) integer shifts
        shifts_r = shifts_i.dot(self.bounds)                    # (S, d) Cartesian shifts
        S        = shifts_r.shape[0]
        d        = self.data.shape[1]

        # Tile 'other' points by all shifts: shape (S * n_other, d)
        O = np.asarray(other.data, float)
        tiled_other = (O[None, :, :] + shifts_r[:, None, :]).reshape(S * other._n_orig, d)

        # Build a plain cKDTree on the tiled 'other' points (no periodicity here)
        tiled_tree = cKDTree(tiled_other, leafsize=other.leafsize)

        # Query sparse distances from *this* base tree to the tiled-other tree
        raw = super(PeriodicCKDTree, self).sparse_distance_matrix(tiled_tree, max_distance, p)

        # Fold back tiled indices to base indices, keeping the minimum distance per (i,j)
        result = {}
        same_obj = (self is other)

        # Identify the index of the zero lattice shift to drop (i,i) when self is other
        zero_shift_idx = None
        zmask = np.all(shifts_i == 0, axis=1)
        if np.any(zmask):
            zero_shift_idx = int(np.nonzero(zmask)[0][0])

        for (i, j_tiled), dist in raw.items():
            # Recover the original j and the image (shift) index
            j = j_tiled % other._n_orig
            img_idx = j_tiled // other._n_orig

            # If querying the same object, exclude the zero-shift self-pair and enforce (i<j)
            if same_obj:
                if (zero_shift_idx is not None) and (img_idx == zero_shift_idx) and (i == j):
                    continue
                if j < i:
                    i, j = j, i

            key = (i, j)
            prev = result.get(key)
            if (prev is None) or (dist < prev):
                result[key] = dist

        return result



"""
Molecule Builder Library

A Python library for constructing molecular structures, supporting diatomic and triatomic molecules.
Molecular geometries are based on experimental data.

Dependencies:
- numpy for numerical operations
- sage_lib for partition management and atomic position handling

Part of the sage_lib framework.
"""
try:
    from ...IO.structure_handling_tools.AtomPosition import AtomPosition
    from .BasePartition import BasePartition
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing AtomPosition: {str(e)}\n")
    del sys

try:
    import numpy as np
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing numpy: {str(e)}\n")
    del sys

class Molecule_builder(BasePartition):
    def __init__(self, file_location: str = None, name: str = None, *args, **kwargs):
        """
        """

        self._diatomic_compounds = {
            'H2':  {'atomLabels': ['H', 'H'],   'atomPositions': [[0, 0, 0], [0, 0,  .62]]},
            'O2':  {'atomLabels': ['O', 'O'],   'atomPositions': [[0, 0, 0], [0, 0, 1.32]]},
            'N2':  {'atomLabels': ['N', 'N'],   'atomPositions': [[0, 0, 0], [0, 0, 1.42]]},
            'F2':  {'atomLabels': ['F', 'F'],   'atomPositions': [[0, 0, 0], [0, 0, 1.14]]},
            'Cl2': {'atomLabels': ['Cl', 'Cl'], 'atomPositions': [[0, 0, 0], [0, 0, 2.04]]},
            'Br2': {'atomLabels': ['Br', 'Br'], 'atomPositions': [[0, 0, 0], [0, 0, 2.40]]},
            'I2':  {'atomLabels': ['I', 'I'],   'atomPositions': [[0, 0, 0], [0, 0, 2.78]]},
            'HF':  {'atomLabels': ['H', 'F'],   'atomPositions': [[0, 0, 0], [0, 0,  .88]]},
            'CO':  {'atomLabels': ['C', 'O'],   'atomPositions': [[0, 0, 0], [0, 0, 1.42]]},
            'NO':  {'atomLabels': ['N', 'O'],   'atomPositions': [[0, 0, 0], [0, 0, 1.37]]}
                            }
                            
        self._triatomic_compounds = {
            'CO2': {'atomLabels': ['C', 'O', 'O'], 
                    'atomPositions': [[0, 0, 0], 
                                    [0, 0, 1.16], 
                                    [0, 0, -1.16]]},
            'H2O': {'atomLabels': ['O', 'H', 'H'], 
                    'atomPositions': [[0, 0, 0], 
                                    [.9584 * np.cos(np.radians(104.5/2)), .9584 * np.sin(np.radians(104.5/2))], 
                                    [.9584 * np.cos(np.radians(104.5/2)), -.9584 * np.cos(np.radians(104.5/2))]]},
            'SO2': {'atomLabels': ['S', 'O', 'O'], 
                    'atomPositions': [[0, 0, 0], 
                                    [1.57 * np.sin(np.radians(119.5/2)), 0, -1.57 * np.cos(np.radians(119.5/2))], 
                                    [-1.57 * np.sin(np.radians(119.5/2)), 0, -1.57 * np.cos(np.radians(119.5/2))]]},
            'O3': {'atomLabels': ['O', 'O', 'O'], 
                'atomPositions': [[0, 0, 0], 
                                    [1.28 * np.sin(np.radians(116.8/2)), 0, -1.28 * np.cos(np.radians(116.8/2))], 
                                    [-1.28 * np.sin(np.radians(116.8/2)), 0, -1.28 * np.cos(np.radians(116.8/2))]]},
            'HCN': {'atomLabels': ['H', 'C', 'N'], 
                    'atomPositions': [[0, 0, 1.20], 
                                    [0, 0, 0], 
                                    [0, 0, -1.16]]},
            'NO2': {'atomLabels': ['N', 'O', 'O'],  # Nitrógeno dióxido
                    'atomPositions': [[0, 0, 0], 
                                    [1.19 * np.sin(np.radians(134.0/2)), 0, -1.19 * np.cos(np.radians(134.0/2))], 
                                    [-1.19 * np.sin(np.radians(134.0/2)), 0, -1.19 * np.cos(np.radians(134.0/2))]]},
            'OCS': {'atomLabels': ['O', 'C', 'S'],  # Carbonyl sulfide
                    'atomPositions': [[0, 0, 1.16],  # O
                                    [0, 0, 0],     # C
                                    [0, 0, -1.19]]},  # S
            'HCO': {'atomLabels': ['H', 'C', 'O'],  # Hidrogenocarbó
                    'atomPositions': [[0, 0, 1.12], 
                                    [0, 0, 0], 
                                    [0, 0, -1.13]]},
            'HNO': {'atomLabels': ['H', 'N', 'O'],  # Ácido nitroso
                    'atomPositions': [[0, 0, 1.07], 
                                    [0, 0, 0], 
                                    [0, 0, -1.16]]},
            'HOCl': {'atomLabels': ['H', 'O', 'Cl'],  # Ácido hipocloroso
                    'atomPositions': [[0, 0, 0], 
                                    [0.98 * np.cos(np.radians(102.5/2)), 0.98 * np.sin(np.radians(102.5/2)), 0], 
                                    [1.47 * np.cos(np.radians(102.5/2)), 1.47 * np.sin(np.radians(102.5/2)), 0]]},
            'NH2': {'atomLabels': ['N', 'H', 'H'],  # Amonia (estado reactivo)
                    'atomPositions': [[0, 0, 0], 
                                    [1.01 * np.cos(np.radians(106.6/2)), 1.01 * np.sin(np.radians(106.6/2)), 0], 
                                    [1.01 * np.cos(np.radians(106.6/2)), -1.01 * np.sin(np.radians(106.6/2)), 0]]},
            'HOO': {'atomLabels': ['H', 'O', 'O'],  # Radical hidroxilo
                    'atomPositions': [[0, 0, 0], 
                                    [0.95 * np.cos(np.radians(119.0/2)), 0.95 * np.sin(np.radians(119.0/2)), 0], 
                                    [0.95 * np.cos(np.radians(119.0/2)), -0.95 * np.sin(np.radians(119.0/2)), 0]]},
            'C3': {'atomLabels': ['C', 'C', 'C'],  # Tricarbono
                'atomPositions': [[0, 0, 0], 
                                    [1.20, 0, 0], 
                                    [2.40, 0, 0]]},
            'LiO2': {'atomLabels': ['Li', 'O', 'O'],  # Superoxido de litio
                    'atomPositions': [[0, 0, 0], 
                                        [1.10 * np.cos(np.radians(117.0/2)), 1.10 * np.sin(np.radians(117.0/2)), 0], 
                                        [1.10 * np.cos(np.radians(117.0/2)), -1.10 * np.sin(np.radians(117.0/2)), 0]]},
            'NaO2': {'atomLabels': ['Na', 'O', 'O'],  # Superoxido de sodio
                    'atomPositions': [[0, 0, 0], 
                                    [1.32 * np.cos(np.radians(117.0/2)), 1.32 * np.sin(np.radians(117.0/2)), 0], 
                                    [1.32 * np.cos(np.radians(117.0/2)), -1.32 * np.sin(np.radians(117.0/2)), 0]]},
            'KO2': {'atomLabels': ['K', 'O', 'O'],  # Superoxido de potasio
                    'atomPositions': [[0, 0, 0], 
                                    [1.47 * np.cos(np.radians(117.0/2)), 1.47 * np.sin(np.radians(117.0/2)), 0], 
                                    [1.47 * np.cos(np.radians(117.0/2)), -1.47 * np.sin(np.radians(117.0/2)), 0]]},
            'MgO2': {'atomLabels': ['Mg', 'O', 'O'],  # Óxido de magnesio
                    'atomPositions': [[0, 0, 0], 
                                    [1.30 * np.cos(np.radians(120.0/2)), 1.30 * np.sin(np.radians(120.0/2)), 0], 
                                    [1.30 * np.cos(np.radians(120.0/2)), -1.30 * np.sin(np.radians(120.0/2)), 0]]},
            'CH2': {'atomLabels': ['C', 'H', 'H'],  # Metilideno
                    'atomPositions': [[0, 0, 0], 
                                    [0.77 * np.cos(np.radians(135.0/2)), 0.77 * np.sin(np.radians(135.0/2)), 0], 
                                    [0.77 * np.cos(np.radians(135.0/2)), -0.77 * np.sin(np.radians(135.0/2)), 0]]},
            'BH2': {'atomLabels': ['B', 'H', 'H'],  # Boro hidruro
                    'atomPositions': [[0, 0, 0], 
                                    [1.18 * np.cos(np.radians(120.0/2)), 1.18 * np.sin(np.radians(120.0/2)), 0], 
                                    [1.18 * np.cos(np.radians(120.0/2)), -1.18 * np.sin(np.radians(120.0/2)), 0]]},
            'AlO': {'atomLabels': ['Al', 'O'],  # Óxido de aluminio (diccionario existente)
                    'atomPositions': [[0, 0, 0], 
                                    [0, 0, 1.61]]}
        }

        super().__init__(*args, **kwargs)

    def build_molecule(self, atomLabels:list, atomPositions:np.array):
        molecule = AtomPosition()

        for al, ap in zip(atomLabels, atomPositions):
            molecule.add_atom(al, ap, [1,1,1])
        
        return molecule

    def build(self, name:str):

        if name in self.diatomic_compounds:
            atomLabels = self.diatomic_compounds[name]['atomLabels']
            atomPositions = self.diatomic_compounds[name]['atomPositions']
            molecule = self.build_molecule(atomLabels, atomPositions)

        if name in self.triatomic_compounds:
            atomLabels = self.triatomic_compounds[name]['atomLabels']
            atomPositions = self.triatomic_compounds[name]['atomPositions']
            molecule = self.build_molecule(atomLabels, atomPositions)

        return molecule

'''
mb = MoleculeBuilder()
mb.build('H2O')
print( mb.atomPositions )
'''